from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Optional

from syntheseus.reaction_prediction.data.reaction_sample import ReactionSample, ReactionType
from syntheseus.reaction_prediction.utils.misc import undictify_bag_of_molecules

from retrochimera.chem.rewrite import RewriteResult


@dataclass(frozen=True)
class TemplateApplicationResult:
    """Result from applying a fixed template on a given product.

    If `results` is `None`, that means the template application result was not recorded, and a
    downstream user should recompute it if needed. If `results` is an empty list, then the template
    was applied, but it did not match.

    The `ground_truth` flag is set if the template was produced by running template extraction on a
    data sample during preprocessing; usually this means that `results` contains a result with the
    ground-truth reactant bag, although in some situations this may not be the case (due to e.g
    stereochemistry issues).
    """

    template_id: int  # Corresponding template ID
    ground_truth: bool = True  # Whether this template was produced by running extraction
    results: Optional[list[RewriteResult]] = None  # Results from applying the template

    @classmethod
    def from_dict(
        cls: type[TemplateApplicationResult], data: dict[str, Any]
    ) -> TemplateApplicationResult:
        if data["results"] is not None:
            for r in data["results"]:
                r["mols"] = undictify_bag_of_molecules(r["mols"])

            data["results"] = [RewriteResult(**r) for r in data["results"]]

        return cls(**data)


@dataclass(frozen=True)
class TemplateReactionSample(ReactionSample):
    """Reaction sample containing results from applying one or many templates."""

    template: Optional[str] = field(default=None, hash=False, compare=False)
    template_application_results: list[TemplateApplicationResult] = field(
        default_factory=lambda: []
    )

    @classmethod
    def from_dict(cls: type[ReactionType], data: dict[str, Any]) -> ReactionType:
        data["template_application_results"] = [
            TemplateApplicationResult.from_dict(d) for d in data["template_application_results"]
        ]

        return super().from_dict(data)



import tempfile
from dataclasses import asdict
from pathlib import Path
from typing import Generator

import numpy as np
import pytest

from retrochimera.data.dataset import DataFold, ReactionDataset
from retrochimera.data.preprocessing import template_classification, template_localization
from retrochimera.data.processed_dataset import ProcessedDataModule
from retrochimera.data.template_reaction_sample import TemplateReactionSample
from retrochimera.encoders.chem import ChemEncoder
from retrochimera.encoders.configs import GNNEncoderConfig
from retrochimera.encoders.featurizers import DGLLifeRewriteFeaturizer
from retrochimera.encoders.gnn import GNNEncoder
from retrochimera.utils.training import preprocess_and_save


def run_preprocessing_with_dummy_model(model, dataset: ReactionDataset):
    with tempfile.TemporaryDirectory() as temp_dir:
        save_path = Path(temp_dir, "data.h5")
        preprocess_and_save(
            save_path=save_path,
            dataset=dataset,
            model=model,
            num_processes=1,
        )

        datamodule = ProcessedDataModule(
            h5_path=save_path, data_loader_kwargs={"collate_fn": lambda x: x, "shuffle": False}
        )

        train_data: list = sum(datamodule.train_dataloader(), [])
        test_data: list = sum(datamodule.test_dataloader(), [])

        assert len(train_data) == dataset.get_num_samples(DataFold.TRAIN)
        assert len(test_data) == dataset.get_num_samples(DataFold.TEST)

        return train_data, test_data


@pytest.fixture
def rulebase_dir(reaction_dataset: ReactionDataset) -> Generator[str, None, None]:
    with tempfile.TemporaryDirectory() as temp_dir:
        reaction_dataset.rulebase.save_to_file(dir=temp_dir)
        yield temp_dir


def test_preprocess_classification(reaction_dataset: ReactionDataset, rulebase_dir: str):
    class DummyModel:
        def __init__(self):
            self.encoder = ChemEncoder(fingerprint_dim=16)

        def preprocess(self, samples: list[TemplateReactionSample], num_processes: int):
            return template_classification.preprocess_samples(
                samples=samples,
                rulebase_dir=rulebase_dir,
                encoder=self.encoder,
                num_processes=num_processes,
            )

    train_data, test_data = run_preprocessing_with_dummy_model(
        model=DummyModel(), dataset=reaction_dataset
    )

    for processed_sample, raw_sample in zip(train_data, reaction_dataset[DataFold.TRAIN]):
        assert processed_sample.target == raw_sample.template_application_results[0].template_id

    for processed_sample, raw_sample in zip(test_data, reaction_dataset[DataFold.TEST]):
        assert processed_sample.target == raw_sample.template_application_results[0].template_id


def test_preprocess_localization(reaction_dataset: ReactionDataset, rulebase_dir: str):
    rulebase = reaction_dataset.rulebase

    class DummyModel:
        def __init__(self):
            input_encoder_config = GNNEncoderConfig()
            rewrite_encoder_config = GNNEncoderConfig(
                featurizer_class="DGLLifeRewriteFeaturizer",
                featurizer_kwargs=DGLLifeRewriteFeaturizer.prepare_kwargs(
                    rewrites=(rule.rxn for rule in rulebase.rules.values())
                ),
            )

            self.input_encoder = GNNEncoder(**asdict(input_encoder_config))
            self.rewrite_encoder = GNNEncoder(**asdict(rewrite_encoder_config))

        def preprocess(self, samples: list[TemplateReactionSample], num_processes: int):
            return template_localization.preprocess_samples(
                samples=samples,
                rulebase_dir=rulebase_dir,
                input_encoder=self.input_encoder,
                num_processes=num_processes,
            )

    train_data, test_data = run_preprocessing_with_dummy_model(
        model=DummyModel(), dataset=reaction_dataset
    )

    for sample in train_data + test_data:
        # Localization target should have the correct shape.
        assert sample.loc_target.shape == (
            rulebase[sample.target].rxn.rdkit_lhs_mol.GetNumAtoms(),
            len(sample.input.atom_features),
        )

        # Localization target for each rewrite node should be a valid distribution.
        assert np.all(sample.loc_target >= 0.0)
        assert np.all(sample.loc_target <= 1.0)
        np.testing.assert_allclose(sample.loc_target.sum(axis=1), 1.0)



"""
MIT License

Copyright (c) 2017-Present OpenNMT

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Original Source: https://github.com/OpenNMT/OpenNMT-py/blob/v3.5.1/onmt/encoders/transformer.py
Modifications:
1. Removed the `parallel_gpu` argument from all four classes: `TransformerDecoderLayerBase`, `TransformerDecoderLayer`, `TransformerDecoderBase`, and `TransformerDecoder`, as Chimera typically uses Pytorch-Lightning for distributed training.
2. Removed the `embeddings` argument from the `TransformerDecoderBase` and `TransformerDecoder` constructors, as it is now built in the `retrochimera.models.smiles_transformer.SmilesTransformerModel` class.
3. Removed two unused classes: `TransformerLMDecoderLayer` and `TransformerLMDecoder`.
3. Refactored the `forward` method of the `TransformerDecoder` class to align with the `retrochimera.models.smiles_transformer.SmilesTransformerModel` class.
"""
from typing import Optional

import torch
import torch.nn as nn

from retrochimera.opennmt.modules.average_attention import AverageAttention
from retrochimera.opennmt.modules.moe import MoE
from retrochimera.opennmt.modules.multi_head_attention import MultiHeadedAttention
from retrochimera.opennmt.modules.position_ffn import ActivationFunction, PositionwiseFeedForward
from retrochimera.opennmt.modules.rmsnorm import RMSNorm


def sequence_mask(
    lengths: torch.Tensor,
    max_len: Optional[int] = None,
) -> torch.Tensor:
    """Creates a boolean mask from sequence lengths."""
    assert max_len >= lengths.max(), f"{max_len} should be equal to {lengths.max()}"
    max_len = max_len or lengths.max()
    return torch.arange(0, max_len, device=lengths.device) >= lengths.unsqueeze(1)


class TransformerDecoderLayerBase(nn.Module):
    def __init__(
        self,
        d_model,
        heads,
        d_ff,
        dropout,
        attention_dropout,
        self_attn_type="scaled_dot",
        max_relative_positions=0,
        relative_positions_buckets=0,
        aan_useffn=False,
        full_context_alignment=False,
        alignment_heads=0,
        pos_ffn_activation_fn=ActivationFunction.relu,
        add_qkvbias=False,
        num_kv=0,
        add_ffnbias=True,
        parallel_residual=False,
        shared_layer_norm=False,
        layer_norm="standard",
        norm_eps=1e-6,
        use_ckpting=[],
        sliding_window=0,
        rotary_interleave=True,
        rotary_theta=1e4,
        rotary_dim=0,
        num_experts=0,
        num_experts_per_tok=2,
    ):
        """
        Args:
            d_model (int): the dimension of keys/values/queries in MultiHeadedAttention,
            also the input size of the first-layer of PositionwiseFeedForward.
            heads (int): the number of heads for MultiHeadedAttention.
            d_ff (int): the second-layer of PositionwiseFeedForward.
            dropout (float): dropout in residual, self-attn(dot) and feed-forward.
            attention_dropout (float): dropout in context_attn  (and self-attn(avg)).
            self_attn_type (string): type of self-attention scaled-dot, flash-scaled-dot, average.
            max_relative_positions (int): max distance between inputs in relative positions representations.
            relative_positions_buckets (int): relative position bias see
            https://github.com/google-research/text-to-text-transfer-transformer.
            aan_useffn (bool): Turn on the FFN layer in the AAN decoder.
            full_context_alignment (bool): whether enable an extra full context decoder forward for alignment.
            alignment_heads (int): number of cross attention heads to use for alignment guiding.
            pos_ffn_activation_fn (ActivationFunction): activation function choice for PositionwiseFeedForward layer
            add_qkvbias (bool): whether to add bias to the Key/Value nn.Linear.
            num_kv (int): number of heads for KV when different vs Q (multiquery).
            add_ffnbias (bool): whether to add bias to the FF nn.Linear.
            parallel_residual (bool): use parallel residual connections in each layer block, as used by the GPT-J and GPT-NeoX models.
            shared_layer_norm (bool): when using parallel residual, share the input and post attention layer norms.
            layer_norm (string): type of layer normalization standard/rms.
            norm_eps (float): layer norm epsilon.
            use_ckpting (List): layers for which we checkpoint for backward.
            sliding_window (int): width of the band mask and KV cache (cf Mistral Model).
            rotary_interleave (bool): interleave the head dimensions when rotary embeddings are applied.
            rotary_theta (float): rotary base theta.
            rotary_dim (int): in some cases the rotary dim is lower than head dim.
            num_experts (int): number of experts for MoE.
            num_experts_per_tok (int): number of experts choice per token.
        """
        super(TransformerDecoderLayerBase, self).__init__()

        if self_attn_type in ["scaled-dot", "scaled-dot-flash"]:
            self.self_attn = MultiHeadedAttention(
                heads,
                d_model,
                dropout=attention_dropout,
                max_relative_positions=max_relative_positions,
                relative_positions_buckets=relative_positions_buckets,
                rotary_interleave=rotary_interleave,
                rotary_theta=rotary_theta,
                rotary_dim=rotary_dim,
                attn_type="self",
                self_attn_type=self_attn_type,
                add_qkvbias=add_qkvbias,
                num_kv=num_kv,
                use_ckpting=use_ckpting,
            )
        elif self_attn_type == "average":
            self.self_attn = AverageAttention(
                d_model, dropout=attention_dropout, aan_useffn=aan_useffn
            )

        if num_experts > 0:
            self.feed_forward = MoE(
                num_experts,
                num_experts_per_tok,
                d_model,
                d_ff,
                dropout,
                pos_ffn_activation_fn,
                add_ffnbias,
                parallel_residual,
                layer_norm,
                norm_eps,
                use_ckpting=use_ckpting,
            )
        else:
            self.feed_forward = PositionwiseFeedForward(
                d_model,
                d_ff,
                dropout,
                pos_ffn_activation_fn,
                add_ffnbias,
                parallel_residual,
                layer_norm,
                norm_eps,
                use_ckpting=use_ckpting,
            )
        self.parallel_residual = parallel_residual
        self.shared_layer_norm = shared_layer_norm
        if layer_norm == "standard":
            self.layer_norm_1 = nn.LayerNorm(d_model, eps=norm_eps)
            if parallel_residual and not shared_layer_norm:
                self.layer_norm_res = nn.LayerNorm(d_model, eps=norm_eps)
        elif layer_norm == "rms":
            self.layer_norm_1 = RMSNorm(d_model, eps=norm_eps)
            if parallel_residual and not shared_layer_norm:
                self.layer_norm_res = RMSNorm(d_model, eps=norm_eps)
        else:
            raise ValueError(f"{layer_norm} layer norm type is not supported")

        self.dropout = nn.Dropout(dropout)
        self.dropout_p = dropout
        self.full_context_alignment = full_context_alignment
        self.alignment_heads = alignment_heads
        self.sliding_window = sliding_window
        self.self_attn_type = self_attn_type

    def forward(self, *args, **kwargs):
        """Extend `_forward` for (possibly) multiple decoder pass:
        Always a default (future masked) decoder forward pass,
        Possibly a second future aware decoder pass for joint learn
        full context alignement, :cite:`garg2019jointly`.

        Args:
            * All arguments of _forward, of which
            with_align (bool): needed to compute attn_align.
            return_attn (bool): to force MHA to return attns.

        Returns:
            layer_out (FloatTensor): shape (batch_size, T, hidden_dim).
            top_attn (FloatTensor): shape (batch_size, T, src_len).
            attn_align (FloatTensor or None): shape (batch_size, T, src_len).
        """
        with_align = kwargs.pop("with_align", False)
        layer_out, attns = self._forward(*args, **kwargs)
        top_attn = None if attns is None else attns[:, 0, :, :].contiguous()
        attn_align = None
        if with_align:
            if self.full_context_alignment:
                # return _, (B, Q_len, K_len)
                _, attns = self._forward(*args, **kwargs, future=True)

            if self.alignment_heads > 0:
                attns = attns[:, : self.alignment_heads, :, :].contiguous()
            # layer average attention across heads, get ``(B, Q, K)``
            # Case 1: no full_context, no align heads -> layer avg baseline
            # Case 2: no full_context, 1 align heads -> guided align
            # Case 3: full_context, 1 align heads -> full cte guided align
            attn_align = attns.mean(dim=1)
        return layer_out, top_attn, attn_align

    def update_dropout(self, dropout, attention_dropout):
        self.self_attn.update_dropout(attention_dropout)
        self.feed_forward.update_dropout(dropout)
        self.dropout.p = dropout

    def _forward(self, *args, **kwargs):
        raise NotImplementedError

    def _compute_dec_mask(self, tgt_pad_mask, future):
        tgt_len = tgt_pad_mask.size(-1)
        if not future:
            # Add triangular future_mask and pad_mask, result mask in (B, T, T).
            future_mask = torch.ones(
                [tgt_len, tgt_len],
                device=tgt_pad_mask.device,
                dtype=torch.uint8,
            )
            future_mask = future_mask.tril_(0)
            if self.sliding_window > 0:
                future_mask = future_mask.triu_(-self.sliding_window)
            future_mask = future_mask.bool()
            future_mask = ~future_mask.view(1, tgt_len, tgt_len)
            # Patch for scaled dot product attention.
            patch_mask = ~torch.all(tgt_pad_mask + future_mask, dim=2, keepdim=True).expand_as(
                tgt_pad_mask + future_mask
            )
            dec_mask = torch.gt(tgt_pad_mask + future_mask, 0)
            dec_mask = torch.logical_and(dec_mask, patch_mask)
        else:
            # Only mask padding, result mask in (B, 1, T).
            dec_mask = tgt_pad_mask
        return dec_mask

    def _forward_self_attn(self, norm_layer_in, dec_mask, step, return_attn=False):
        if self.self_attn_type in ["scaled-dot", "scaled-dot-flash"]:
            return self.self_attn(
                norm_layer_in,
                norm_layer_in,
                norm_layer_in,
                mask=dec_mask,
                sliding_window=self.sliding_window,
                step=step,
                return_attn=return_attn,
            )
        elif self.self_attn_type == "average":
            return self.self_attn(norm_layer_in, mask=dec_mask, step=step)
        else:
            raise ValueError(f"self attention {type(self.self_attn)} not supported")


class TransformerDecoderLayer(TransformerDecoderLayerBase):
    """Transformer Decoder layer block in Pre-Norm style.
    Pre-Norm style is an improvement w.r.t. Original paper's Post-Norm style,
    providing better converge speed and performance. This is also the actual
    implementation in tensor2tensor and also avalable in fairseq.
    See https://tunz.kr/post/4 and :cite:`DeeperTransformer`.
    """

    def __init__(
        self,
        d_model,
        heads,
        d_ff,
        dropout,
        attention_dropout,
        self_attn_type="scaled-dot",
        max_relative_positions=0,
        relative_positions_buckets=0,
        aan_useffn=False,
        full_context_alignment=False,
        alignment_heads=0,
        pos_ffn_activation_fn=ActivationFunction.relu,
        add_qkvbias=False,
        num_kv=0,
        add_ffnbias=True,
        parallel_residual=False,
        shared_layer_norm=False,
        layer_norm="standard",
        norm_eps=1e-6,
        use_ckpting=[],
        sliding_window=0,
        rotary_interleave=True,
        rotary_theta=1e4,
        rotary_dim=0,
        num_experts=0,
        num_experts_per_tok=2,
    ):
        """
        Args:
            See TransformerDecoderLayerBase
        """
        super(TransformerDecoderLayer, self).__init__(
            d_model,
            heads,
            d_ff,
            dropout,
            attention_dropout,
            self_attn_type,
            max_relative_positions,
            relative_positions_buckets,
            aan_useffn,
            full_context_alignment,
            alignment_heads,
            pos_ffn_activation_fn=pos_ffn_activation_fn,
            add_qkvbias=add_qkvbias,
            num_kv=num_kv,
            add_ffnbias=add_ffnbias,
            parallel_residual=parallel_residual,
            shared_layer_norm=shared_layer_norm,
            layer_norm=layer_norm,
            norm_eps=norm_eps,
            use_ckpting=use_ckpting,
            sliding_window=sliding_window,
            rotary_interleave=rotary_interleave,
            rotary_theta=rotary_theta,
            rotary_dim=rotary_dim,
            num_experts=num_experts,
            num_experts_per_tok=num_experts_per_tok,
        )
        self.context_attn = MultiHeadedAttention(
            heads,
            d_model,
            dropout=attention_dropout,
            attn_type="context",
            self_attn_type=self.self_attn_type,
            add_qkvbias=add_qkvbias,
            num_kv=num_kv,
            use_ckpting=use_ckpting,
        )
        if layer_norm == "standard":
            self.layer_norm_2 = nn.LayerNorm(d_model, eps=norm_eps)
        elif layer_norm == "rms":
            self.layer_norm_2 = RMSNorm(d_model, eps=norm_eps)
        else:
            raise ValueError(f"{layer_norm} layer norm type is not supported")

    def update_dropout(self, dropout, attention_dropout):
        super(TransformerDecoderLayer, self).update_dropout(dropout, attention_dropout)
        self.context_attn.update_dropout(attention_dropout)

    def _forward(
        self,
        layer_in,
        enc_out,
        src_pad_mask,
        tgt_pad_mask,
        step=None,
        future=False,
        return_attn=False,
    ):
        """A naive forward pass for transformer decoder.

        # T: could be 1 in the case of stepwise decoding or tgt_len

        Args:
            layer_in (FloatTensor): ``(batch_size, T, hidden_dim)``
            enc_out (FloatTensor): ``(batch_size, src_len, hidden_dim)``
            src_pad_mask (bool): ``(batch_size, 1, src_len)``
            tgt_pad_mask (bool): ``(batch_size, 1, T)``
            step (int or None): stepwise decoding counter
            future (bool): If set True, do not apply future_mask.
            return_attn (bool) : if set True requires attns output

        Returns:
            (FloatTensor, FloatTensor):

            * layer_out ``(batch_size, T, hidden_dim)``
            * attns ``(batch_size, head, T, src_len)``

        """
        dec_mask = None
        src_pad_mask = src_pad_mask.unsqueeze(1)  # [B,1,1,slen]

        if layer_in.size(1) > 1:
            # masking is necessary when sequence length is greater than one
            dec_mask = self._compute_dec_mask(tgt_pad_mask, future)
            dec_mask = dec_mask.unsqueeze(1)
            dec_mask = dec_mask.expand(-1, -1, dec_mask.size(3), -1)
            src_pad_mask = src_pad_mask.expand(-1, -1, dec_mask.size(3), -1)
            # mask now are (batch x 1 x tlen x s or t len)
            # 1 = heads to be expanded in MHA

        norm_layer_in = self.layer_norm_1(layer_in)

        self_attn, _ = self._forward_self_attn(
            norm_layer_in, dec_mask, step, return_attn=return_attn
        )
        if self.dropout_p > 0:
            self_attn = self.dropout(self_attn)
        if self.parallel_residual:
            ctx_attn, attns = self.context_attn(
                enc_out,
                enc_out,
                norm_layer_in,
                mask=src_pad_mask,
                return_attn=return_attn,
            )
            # feed_forward applies residual, so we remove and apply residual with un-normed
            layer_out = (
                self.feed_forward(norm_layer_in) - norm_layer_in + layer_in + self_attn + ctx_attn
            )
        else:
            query = self_attn + layer_in
            norm_query = self.layer_norm_2(query)
            ctx_attn, attns = self.context_attn(
                enc_out, enc_out, norm_query, mask=src_pad_mask, return_attn=return_attn
            )
            if self.dropout_p > 0:
                ctx_attn = self.dropout(ctx_attn)
            layer_out = self.feed_forward(ctx_attn + query)

        return layer_out, attns


class TransformerDecoderBase(nn.Module):
    def __init__(self, d_model, copy_attn, alignment_layer, layer_norm, norm_eps):
        super(TransformerDecoderBase, self).__init__()

        # Decoder State
        self.state = {}

        # previously, there was a GlobalAttention module here for copy
        # attention. But it was never actually used -- the "copy" attention
        # just reuses the context attention.
        self._copy = copy_attn
        if layer_norm == "standard":
            self.layer_norm = nn.LayerNorm(d_model, eps=norm_eps)
        elif layer_norm == "rms":
            self.layer_norm = RMSNorm(d_model, eps=norm_eps)
        else:
            raise ValueError(f"{layer_norm} layer norm type is not supported")

        self.alignment_layer = alignment_layer

    def init_state(self, src, enc_out, enc_final_hs):
        """Initialize decoder state."""
        self.state["src"] = src  # torch.Tensor: (padded_src_len, batch_size, 1)

    def map_state(self, fn, only_map_src=False):
        if self.state["src"] is not None:
            self.state["src"] = fn(self.state["src"], 1)

        if not only_map_src:
            for layer in self.transformer_layers:
                if hasattr(layer, "context_attn"):
                    if layer.context_attn.layer_cache[1]["keys"].numel() != 0:
                        x = fn(layer.context_attn.layer_cache[1]["keys"], 0)
                        y = fn(layer.context_attn.layer_cache[1]["values"], 0)
                        layer.context_attn.layer_cache = True, {"keys": x, "values": y}
                if isinstance(layer.self_attn, AverageAttention):
                    if layer.self_attn.layer_cache[1]["prev_g"].numel() != 0:
                        x = fn(layer.self_attn.layer_cache[1]["prev_g"], 0)
                        layer.self_attn.layer_cache = True, {"prev_g": x}
                else:
                    if layer.self_attn.layer_cache[1]["keys"].numel() != 0:
                        x = fn(layer.self_attn.layer_cache[1]["keys"], 0)
                        y = fn(layer.self_attn.layer_cache[1]["values"], 0)
                        if layer.self_attn.layer_cache[1].get("key_pad_mask", None) is not None:
                            z = fn(layer.self_attn.layer_cache[1]["key_pad_mask"], 0)
                        else:
                            z = None
                        layer.self_attn.layer_cache = True, {
                            "keys": x,
                            "values": y,
                            "key_pad_mask": z,
                        }

    def detach_state(self):
        raise NotImplementedError

    def forward(self, *args, **kwargs):
        raise NotImplementedError

    def update_dropout(self, dropout, attention_dropout):
        for layer in self.transformer_layers:
            layer.update_dropout(dropout, attention_dropout)


class TransformerDecoder(TransformerDecoderBase):
    """The Transformer decoder from "Attention is All You Need".
    :cite:`DBLP:journals/corr/VaswaniSPUJGKP17`

    Args:
        num_layers (int): number of decoder layers.
        d_model (int): size of the model.
        heads (int): number of heads.
        d_ff (int): size of the inner FF layer.
        copy_attn (bool): if using a separate copy attention.
        self_attn_type (str): type of self-attention scaled-dot, scaled-dot-flash, average.
        dropout (float): dropout in residual, self-attn(dot) and feed-forward.
        attention_dropout (float): dropout in context_attn (and self-attn(avg)).
        max_relative_positions (int): enable relative position encoding: -1 to enable Rotary Embeddings,
        > 0 (e.g., 16, 32) to use max distance between inputs in relative positions representations
        relative_positions_buckets (int): number of buckets when using relative position bias.
        aan_useffn (bool): turn on the FFN layer in the AAN decoder.
        full_context_alignment (bool): whether enable an extra full context decoder forward for alignment.
        alignment_layer (int): N° Layer to supervise with for alignment guiding.
        alignment_heads (int): N. of cross attention heads to use for alignment guiding.
        pos_ffn_activation_fn (str): activation function choice for PositionwiseFeedForward layer.
        add_qkvbias (bool): whether to add bias to the Key/Value nn.Linear.
        num_kv (int): number of heads for KV when different vs Q (multiquery).
        add_ffnbias (bool): whether to add bias to the FF nn.Linear.
        parallel_residual (bool): Use parallel residual connections in each layer block, as used by the GPT-J and GPT-NeoX models.
        shared_layer_norm (bool): When using parallel residual, share the input and post attention layer norms.
        layer_norm (string): type of layer normalization standard/rms.
        norm_eps (float): layer norm epsilon.
        use_ckpting (List): layers for which we checkpoint for backward.
        sliding_window (int): Width of the band mask and KV cache (cf Mistral Model).
        rotary_interleave (bool): Interleave the head dimensions when rotary embeddings are applied.
        rotary_theta (int): rotary base theta.
        rotary_dim (int): in some cases the rotary dim is lower than head dim.
        num_experts (int): Number of experts for MoE.
        num_experts_per_tok (int): Number of experts choice per token.
    """

    def __init__(
        self,
        num_layers: int = 6,
        d_model: int = 256,
        heads: int = 8,
        d_ff: int = 2048,
        copy_attn: bool = False,
        self_attn_type: str = "scaled-dot",
        dropout: float = 0.1,
        attention_dropout: float = 0.1,
        max_relative_positions: int = 0,
        relative_positions_buckets: int = 0,
        aan_useffn: bool = False,
        full_context_alignment: bool = False,
        alignment_layer: int = 0,
        alignment_heads: int = 0,
        pos_ffn_activation_fn: str = ActivationFunction.relu,
        add_qkvbias: bool = False,
        num_kv: int = 0,
        add_ffnbias: bool = True,
        parallel_residual: bool = False,
        shared_layer_norm: bool = False,
        layer_norm: str = "standard",
        norm_eps: float = 1e-6,
        use_ckpting: list = [],
        sliding_window: int = 0,
        rotary_interleave: bool = True,
        rotary_theta: float = 1e4,
        rotary_dim: int = 0,
        num_experts: int = 0,
        num_experts_per_tok: int = 2,
    ):
        super(TransformerDecoder, self).__init__(
            d_model, copy_attn, alignment_layer, layer_norm, norm_eps
        )
        self.transformer_layers = nn.ModuleList(
            [
                TransformerDecoderLayer(
                    d_model,
                    heads,
                    d_ff,
                    dropout,
                    attention_dropout,
                    self_attn_type=self_attn_type,
                    max_relative_positions=max_relative_positions,
                    relative_positions_buckets=relative_positions_buckets,
                    aan_useffn=aan_useffn,
                    full_context_alignment=full_context_alignment,
                    alignment_heads=alignment_heads,
                    pos_ffn_activation_fn=pos_ffn_activation_fn,
                    add_qkvbias=add_qkvbias,
                    num_kv=num_kv,
                    add_ffnbias=add_ffnbias,
                    parallel_residual=parallel_residual,
                    shared_layer_norm=shared_layer_norm,
                    layer_norm=layer_norm,
                    norm_eps=norm_eps,
                    use_ckpting=use_ckpting,
                    sliding_window=sliding_window,
                    rotary_interleave=rotary_interleave,
                    rotary_theta=rotary_theta,
                    rotary_dim=rotary_dim,
                    num_experts=num_experts,
                    num_experts_per_tok=num_experts_per_tok,
                )
                for i in range(num_layers)
            ]
        )

    def detach_state(self):
        self.state["src"] = self.state["src"].detach()

    def forward(
        self,
        tgt: torch.Tensor,
        tgt_key_padding_mask: torch.Tensor,
        enc_out: torch.Tensor,
        enc_key_padding_mask: torch.Tensor,
        step: Optional[int] = None,
        **kwargs,
    ):
        """
        During training, step is always None, when decoding, step increases.

        Args:
        tgt (torch.Tensor): shape (batch_size, padded_tgt_len/1, hidden_dim).
        tgt_key_padding_mask (torch.Tensor, bool: 0/1): shape (batch_size, padded_tgt_len/1).
        enc_out (torch.Tensor): encoder output (batch_size, padded_src_len, hidden_dim).
        enc_key_padding_mask (torch.Tensor, bool: 0/1): encoder padding mask (batch_size, padded_src_len).

        Returns:
        dec_out (torch.Tensor): decoder output (batch_size, padded_tgt_len/1, hidden_dim).
        attns (dict): attention weights.

        Note: padded_tgt_len is 1 when using kv cache during inference.

        """

        assert enc_out is not None
        if step == 0:
            self._init_cache(enc_out)
        elif step is None:
            for layer in self.transformer_layers:
                if isinstance(layer.self_attn, AverageAttention):
                    layer.self_attn.layer_cache = False, {"prev_g": torch.tensor([])}
                else:
                    layer.self_attn.layer_cache = (
                        False,
                        {"keys": torch.tensor([]), "values": torch.tensor([])},
                    )
                layer.context_attn.layer_cache = (
                    False,
                    {"keys": torch.tensor([]), "values": torch.tensor([])},
                )

        dec_out = tgt
        src_len = enc_key_padding_mask.eq(0).sum(dim=1).long()  # shape: (batch_size,)
        src_max_len = self.state["src"].shape[
            0
        ]  # self.state["src"] shape: (padded_src_len, batch_size, 1)

        src_pad_mask = sequence_mask(src_len, src_max_len).unsqueeze(
            1
        )  # shape: (batch_size, 1, padded_src_len)
        tgt_pad_mask = tgt_key_padding_mask.unsqueeze(1)  # shape: (batch_size, 1, padded_tgt_len/1)

        with_align = kwargs.pop("with_align", False)
        return_attn = with_align or self._copy or kwargs.pop("return_attn", False)
        attn_aligns = []

        for layer in self.transformer_layers:
            dec_out, attn, attn_align = layer(
                dec_out,
                enc_out,
                src_pad_mask,
                tgt_pad_mask,
                step=step,
                with_align=with_align,
                return_attn=return_attn,
            )
            if attn_align is not None:
                attn_aligns.append(attn_align)

        dec_out = self.layer_norm(dec_out)  # shape: (batch_size, padded_tgt_len/1, hidden_dim)

        attns = {
            "std": attn
        }  # shape of attn: (batch_size, heads, padded_tgt_len/1, padded_src_len)
        if self._copy:
            attns["copy"] = attn
        if with_align:
            attns["align"] = attn_aligns[self.alignment_layer]  # `(B, Q, K)`
            # attns["align"] = torch.stack(attn_aligns, 0).mean(0)  # All avg

        return dec_out, attns

    def _init_cache(self, enc_out: torch.Tensor):
        """Initialize the cache for the transformer layers.

        Args:
        enc_out (torch.Tensor): encoder output (batch_size, padded_src_len, hidden_dim).

        """
        batch_size = enc_out.size(0)
        depth = enc_out.size(-1)

        for layer in self.transformer_layers:
            # first value set to True triggered by the beginning of decoding
            # layer_cache becomes active in the MultiHeadedAttention fwd
            layer.context_attn.layer_cache = (
                True,
                {
                    "keys": torch.tensor([], device=enc_out.device),
                    "values": torch.tensor([], device=enc_out.device),
                },
            )
            if isinstance(layer.self_attn, AverageAttention):
                layer.self_attn.layer_cache = True, {
                    "prev_g": torch.zeros((batch_size, 1, depth), device=enc_out.device).to(
                        enc_out.dtype
                    )
                }
            else:
                layer.self_attn.layer_cache = (
                    True,
                    {
                        "keys": torch.tensor([], device=enc_out.device),
                        "values": torch.tensor([], device=enc_out.device),
                    },
                )
                if hasattr(layer.self_attn, "rope"):
                    layer.self_attn.rope = layer.self_attn.rope.to(enc_out.device)
                    layer.self_attn.cos = layer.self_attn.cos.to(enc_out.device)
                    layer.self_attn.sin = layer.self_attn.sin.to(enc_out.device)



"""
MIT License

Copyright (c) 2017-Present OpenNMT

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Original Source: https://github.com/OpenNMT/OpenNMT-py/blob/v3.5.1/onmt/translate/beam_search.py
Modifications:
1. Simplified the structure by removing the `BeamSearchLM` and `GNMTGlobalScorer` class, and merged the `BeamSearchBase` class into the `BeamSearch` class.
2. Introduced the `customised_beam_search` attribute and corresponding logic to the `BeamSearch` class, enabling optimized beam search for retrosynthesis prediction.
"""
import torch

from retrochimera.opennmt.decode.decoder_strategy import DecodeStrategy


class BeamSearch(DecodeStrategy):
    """Beam search for seq2seq/encoder-decoder models.

    Note that the attributes list is not exhaustive. Rather, it highlights
    tensors to document their shape. (Since the state variables' "batch"
    size decreases as beams finish, we denote this axis with a B rather than
    ``batch_size``).

    Args:
        pad (int): See base.
        bos (int): See base.
        eos (int): See base.
        unk (int): See base.
        batch_size (int): See base.
        beam_size (int): Number of beams to use (see base ``parallel_paths``).
        n_best (int): Don't stop until at least this many beams have reached EOS.
        min_length (int): See base.
        max_length (int): See base.
        block_ngram_repeat (int): See base.
        exclusion_tokens (set[int]): See base.
        ratio (float).
        ban_unk_token (bool): See base.
        return_attention (bool): See base.

    Attributes:
        top_beam_finished (ByteTensor): Shape (batch_size,).
        _batch_offset (LongTensor): Shape (batch_size,).
        _beam_offset (LongTensor): Shape (batch_size * beam_size,).
        alive_seq (LongTensor): See base.
        topk_log_probs (FloatTensor): Shape (batch_size, beam_size,). These are the scores used for the topk operation.
        memory_lengths (LongTensor): Lengths of encodings. Used for masking attentions.
        select_indices (LongTensor or NoneType): Shape (batch_size * beam_size,). This is just a flat view of the ``_batch_index``.
        topk_scores (FloatTensor): Shape (batch_size, beam_size)``. These are the scores a sequence will receive if it finishes.
        topk_ids (LongTensor): Shape ``(B, beam_size)``. These are the word indices of the topk predictions.
        _batch_index (LongTensor): Shape ``(B, beam_size)``.
        hypotheses (list[list[Tuple[Tensor]]]): Contains a tuple
            of score (float), sequence (long), and attention (float or None).
    """

    def __init__(
        self,
        pad: int = 1,
        bos: int = 2,
        eos: int = 3,
        unk: int = 0,
        batch_size: int = 32,
        beam_size: int = 10,
        n_best: int = 10,
        min_length: int = 0,
        max_length: int = 512,
        block_ngram_repeat: int = 0,
        exclusion_tokens: set[int] = set(),
        ratio: float = 0,
        ban_unk_token: bool = False,
        return_attention: bool = False,
        customised_beam_search: bool = False,
    ):
        super(BeamSearch, self).__init__(
            pad,
            bos,
            eos,
            unk,
            batch_size,
            beam_size,
            min_length,
            max_length,
            block_ngram_repeat,
            exclusion_tokens,
            ban_unk_token,
            return_attention,
        )

        # beam parameters
        self.beam_size = beam_size
        self.n_best = n_best
        self.ratio = ratio

        # beam state
        self.top_beam_finished = torch.zeros([batch_size], dtype=torch.uint8)
        # BoolTensor was introduced in pytorch 1.2
        try:
            self.top_beam_finished = self.top_beam_finished.bool()
        except AttributeError:
            pass
        self._batch_offset = torch.arange(batch_size, dtype=torch.long)

        self.select_indices = None
        self.done = False
        # "global state" of the old beam
        self.memory_lengths = None

        self.customised_beam_search = customised_beam_search

    def initialize(self, memory_bank, src_lengths, src_map=None, device=None, target_prefix=None):
        """Initialize for decoding.
        Repeat src objects `beam_size` times.

        Args:
            memory_bank (FloatTensor): The memory bank to perform attention. shape (padded_src_len, batch_size, hidden_dim).
            src_lengths (LongTensor): The length of each source sequence. shape (batch_size,).
        """

        (fn_map_state, memory_bank, src_map, target_prefix) = self.initialize_tile(
            memory_bank, src_lengths, src_map, target_prefix
        )
        # shape of memory_bank: (padded_src_len, batch_size * beam_size, hidden_dim)
        # shape of self.memory_lengths: (batch_size * beam_size,)

        if device is None:
            device = self.get_device_from_memory_bank(memory_bank)

        self.initialize_(memory_bank, self.memory_lengths, src_map, device, target_prefix)

        return fn_map_state, memory_bank, self.memory_lengths, src_map

    def initialize_(self, memory_bank, memory_lengths, src_map, device, target_prefix):
        """Initialize for decoding.

        Args:
            memory_bank (FloatTensor): The memory bank to perform attention. shape (padded_src_len, batch_size * beam_size, hidden_dim).
            memory_lengths (LongTensor): The length of each source sequence. shape (batch_size * beam_size,).
        """
        super(BeamSearch, self).initialize(
            memory_bank, memory_lengths, src_map, device, target_prefix
        )

        self.best_scores = torch.full(
            [self.batch_size], -1e10, dtype=torch.float, device=device
        )  # only when ratio > 0
        self._beam_offset = torch.arange(
            0,
            self.batch_size * self.beam_size,
            step=self.beam_size,
            dtype=torch.long,
            device=device,
        )  # shape: (batch_size,)
        self.topk_log_probs = (
            torch.tensor([0.0] + [float("-inf")] * (self.beam_size - 1), device=device)
            .repeat(self.batch_size)
            .reshape(self.batch_size, self.beam_size)
        )
        # buffers for the topk scores and 'backpointer'
        self.topk_scores = torch.empty(
            (self.batch_size, self.beam_size), dtype=torch.float, device=device
        )
        self.topk_ids = torch.empty(
            (self.batch_size, self.beam_size), dtype=torch.long, device=device
        )
        self._batch_index = torch.empty(
            [self.batch_size, self.beam_size], dtype=torch.long, device=device
        )

    @property
    def current_predictions(self) -> torch.Tensor:
        return self.alive_seq[:, -1]

    @property
    def batch_offset(self):
        return self._batch_offset

    def _pick(self, log_probs, out=None):
        """Take a token pick decision for a step.

        Args:
            log_probs (FloatTensor): (B * beam_size, vocab_size)
            out (Tensor, LongTensor): output buffers to reuse, optional.

        Returns:
            topk_scores (FloatTensor): (B, beam_size)
            topk_ids (LongTensor): (B, beam_size)
        """
        vocab_size = log_probs.size(-1)
        # maybe fix some prediction at this step by modifying log_probs
        log_probs = self.target_prefixing(log_probs)

        # Flatten probs into a list of possibilities.
        curr_scores = log_probs.reshape(-1, self.beam_size * vocab_size)
        if out is not None:
            torch.topk(curr_scores, self.beam_size, dim=-1, out=out)
            return
        topk_scores, topk_ids = torch.topk(curr_scores, self.beam_size, dim=-1)
        return topk_scores, topk_ids

    def update_finished(self):
        # Penalize beams that finished.
        _B_old = self.topk_log_probs.shape[0]
        step = self.alive_seq.shape[-1]  # 1 greater than the step in advance

        if not self.customised_beam_search:
            self.topk_log_probs.masked_fill_(
                self.is_finished, -1e10
            )  # shape: (batch_size, beam_size)

        # on real data (newstest2017) with the pretrained transformer,
        # it's faster to not move this back to the original device
        self.is_finished = self.is_finished.to("cpu")
        self.top_beam_finished |= self.is_finished[:, 0].eq(1)
        predictions = self.alive_seq.view(_B_old, self.beam_size, step)
        attention = (
            self.alive_attn.view(step - 1, _B_old, self.beam_size, self.alive_attn.size(-1))
            if self.alive_attn is not None
            else None
        )
        non_finished_batch = []
        for i in range(self.is_finished.size(0)):  # Batch level
            b = self._batch_offset[i]
            finished_hyp = self.is_finished[i].nonzero(as_tuple=False).view(-1)
            # Store finished hypotheses for this batch.
            for j in finished_hyp:  # Beam level: finished beam j in batch i
                if self.ratio > 0:
                    s = self.topk_scores[i, j] / (step + 1)
                    if self.best_scores[b] < s:
                        self.best_scores[b] = s
                if not self.customised_beam_search:
                    self.hypotheses[b].append(
                        (
                            self.topk_scores[i, j],
                            predictions[i, j, 1:],  # Ignore start_token.
                            attention[:, i, j, : self.memory_lengths[i]]
                            if attention is not None
                            else None,
                        )
                    )
                else:
                    if predictions[i, j, 1:].size(-1) == 0 or predictions[i, j, -2] != self.eos:
                        self.hypotheses[b].append(
                            (
                                self.topk_scores[i, j],
                                predictions[i, j, 1:],  # Ignore start_token.
                                attention[:, i, j, : self.memory_lengths[i]]
                                if attention is not None
                                else None,
                            )
                        )
            # End condition is the top beam finished and we can return
            # n_best hypotheses.
            if self.ratio > 0:
                pred_len = self.memory_lengths[i] * self.ratio
                finish_flag = (
                    (self.topk_scores[i, 0] / pred_len) <= self.best_scores[b]
                ) or self.is_finished[i].all()
            else:
                if not self.customised_beam_search:
                    finish_flag = self.top_beam_finished[i] != 0
                else:
                    finish_flag = self.is_finished[i].all()  # shape: (beam_size,)

            if finish_flag and len(self.hypotheses[b]) >= self.n_best:
                best_hyp = sorted(self.hypotheses[b], key=lambda x: x[0], reverse=True)
                for n, (score, pred, attn) in enumerate(best_hyp):
                    if n >= self.n_best:
                        break
                    self.scores[b].append(score)
                    self.predictions[b].append(pred)  # ``(batch, n_best,)``
                    self.attention[b].append(attn if attn is not None else [])
            else:
                non_finished_batch.append(i)
        non_finished = torch.tensor(non_finished_batch)
        # If all sentences are translated, no need to go further.
        if len(non_finished) == 0:
            self.done = True
            return

        _B_new = non_finished.shape[0]
        self.remove_finished_batches(_B_new, _B_old, non_finished, predictions, attention, step)

    def remove_finished_batches(self, _B_new, _B_old, non_finished, predictions, attention, step):
        # Remove finished batches for the next step.
        self.top_beam_finished = self.top_beam_finished.index_select(0, non_finished)
        self._batch_offset = self._batch_offset.index_select(0, non_finished)
        non_finished = non_finished.to(self.topk_ids.device)
        self.topk_log_probs = self.topk_log_probs.index_select(0, non_finished)
        self._batch_index = self._batch_index.index_select(0, non_finished)
        self.select_indices = self._batch_index.view(_B_new * self.beam_size)
        self.alive_seq = predictions.index_select(0, non_finished).view(-1, self.alive_seq.size(-1))
        self.topk_scores = self.topk_scores.index_select(0, non_finished)
        self.topk_ids = self.topk_ids.index_select(0, non_finished)
        self.maybe_update_target_prefix(self.select_indices)
        if self.alive_attn is not None:
            inp_seq_len = self.alive_attn.size(-1)
            self.alive_attn = attention.index_select(1, non_finished).view(
                step - 1, _B_new * self.beam_size, inp_seq_len
            )

    def advance(self, log_probs, attn):
        vocab_size = log_probs.size(-1)

        # using integer division to get an integer _B without casting
        _B = log_probs.shape[0] // self.beam_size

        # force the output to be longer than self.min_length
        step = len(self)
        self.ensure_min_length(log_probs)
        self.ensure_unk_removed(log_probs)

        # Multiply probs by the beam probability.
        log_probs += self.topk_log_probs.view(_B * self.beam_size, 1)

        # if the sequence ends now, then the penalty is the current
        # length + 1, to include the EOS token
        length_penalty = 1.0

        curr_scores = log_probs / length_penalty

        # Avoid any direction that would repeat unwanted ngrams
        self.block_ngram_repeats(curr_scores)

        # Pick up candidate token by curr_scores
        self._pick(curr_scores, out=(self.topk_scores, self.topk_ids))

        # Recover log probs.
        # Length penalty is just a scalar. It doesn't matter if it's applied
        # before or after the topk.
        torch.mul(self.topk_scores, length_penalty, out=self.topk_log_probs)

        # Resolve beam origin and map to batch index flat representation.
        self._batch_index = self.topk_ids // vocab_size
        self._batch_index += self._beam_offset[:_B].unsqueeze(1)
        self.select_indices = self._batch_index.view(_B * self.beam_size)
        self.topk_ids.fmod_(vocab_size)  # resolve true word ids

        # Append last prediction.
        self.alive_seq = torch.cat(
            [
                self.alive_seq.index_select(0, self.select_indices),
                self.topk_ids.view(_B * self.beam_size, 1),
            ],
            -1,
        )  # shape: (B * beam_size, step + 1)

        self.maybe_update_forbidden_tokens()

        if self.return_attention:
            current_attn = attn.index_select(1, self.select_indices)
            if step == 1:
                self.alive_attn = current_attn
                # update global state (step == 1)

            else:
                self.alive_attn = self.alive_attn.index_select(1, self.select_indices)
                self.alive_attn = torch.cat([self.alive_attn, current_attn], 0)
                # update global state (step > 1)

        self.is_finished = self.topk_ids.eq(self.eos)  # shape: (B, beam_size)
        self.ensure_max_length()



import re
from collections import defaultdict
from pathlib import Path
from typing import Optional

from syntheseus.reaction_prediction.data.dataset import DataFold, DiskReactionDataset

from retrochimera.data.smiles_reaction_sample import SmilesReactionSample
from retrochimera.utils.logging import get_logger

# Special tokens
DEFAULT_BEGIN_TOKEN = "<BOS>"
DEFAULT_END_TOKEN = "<EOS>"
DEFAULT_PAD_TOKEN = "<PAD>"
DEFAULT_UNK_TOKEN = "<UNK>"
PATTERN = "(\[[^\]]+]|Br?|Cl?|N|O|S|P|F|I|b|c|n|o|s|p|\(|\)|\.|=|#|-|\+|\\\\|\/|:|~|@|\?|>|\*|\$|\%[0-9]{2}|[0-9])"
REGEX = re.compile(PATTERN)

logger = get_logger(__name__)


class Tokenizer:
    """Tokenizer class for SMILES strings.

    To use it, first create the tokenizer from an existing vocab file or a SMILES file.
    """

    def __init__(
        self,
        vocab: list[str],
        regex: re.Pattern = REGEX,
        begin_token: str = DEFAULT_BEGIN_TOKEN,
        end_token: str = DEFAULT_END_TOKEN,
        pad_token: str = DEFAULT_PAD_TOKEN,
        unk_token: str = DEFAULT_UNK_TOKEN,
    ):
        """Initialize a tokenizer.
        Args:
            vocab: Vocabulary for tokenizer, typically should include the above four special tokens
            regex: Regex object for tokenizing
            begin_token: Token to use at start of each sequence
            end_token: Token to use at end of each sequence
            pad_token: Token to use when padding batches of sequences
            unk_token: Token to use for tokens which are not in the vocabulary
        """
        self.begin_token = begin_token
        self.end_token = end_token
        self.pad_token = pad_token
        self.unk_token = unk_token

        vocab_dict = {
            begin_token: 0,
            end_token: 1,
            pad_token: 2,
            unk_token: 3,
        }
        for token in vocab:
            vocab_dict.setdefault(token, len(vocab_dict))

        self.vocab_dict = vocab_dict  # token -> id
        self.decode_vocab_dict = {i: t for t, i in self.vocab_dict.items()}  # id -> token
        self.regex = regex
        self.begin_token_id = self.vocab_dict[begin_token]
        self.end_token_id = self.vocab_dict[end_token]
        self.pad_token_id = self.vocab_dict[pad_token]
        self.unk_token_id = self.vocab_dict[unk_token]
        self.met_unk_tokens_dict: dict[str, int] = defaultdict(int)

        self.log_vocab_info()

    def log_vocab_info(self) -> None:
        logger.info(f"Vocabulary size: {len(self.vocab_dict)}")
        logger.info(f"Vocabulary: {self.vocab_dict}")

    @staticmethod
    def from_vocab_file(
        vocab_path: str,
        regex: re.Pattern = REGEX,
        begin_token: str = DEFAULT_BEGIN_TOKEN,
        end_token: str = DEFAULT_END_TOKEN,
        pad_token: str = DEFAULT_PAD_TOKEN,
        unk_token: str = DEFAULT_UNK_TOKEN,
    ):
        """Build a tokenizer from a vocab file and regex.
        Args:
            vocab_path: Path to vocab file
            regex: Regex object for tokenizing
        Returns:
            Tokenizer object
        """
        text = Path(vocab_path).read_text()
        tokens = text.split("\n")
        tokens = [t for t in tokens if t is not None and t != ""]

        tokenizer = Tokenizer(
            tokens,
            regex,
            begin_token=begin_token,
            end_token=end_token,
            pad_token=pad_token,
            unk_token=unk_token,
        )
        return tokenizer

    @staticmethod
    def from_smiles(
        smiles: list[str],
        regex: re.Pattern = REGEX,
        begin_token: str = DEFAULT_BEGIN_TOKEN,
        end_token: str = DEFAULT_END_TOKEN,
        pad_token: str = DEFAULT_PAD_TOKEN,
        unk_token: str = DEFAULT_UNK_TOKEN,
    ):
        """Build a tokenizer from smiles strings and regex.
        Args:
            smiles: SMILES strings to use to build vocabulary
            regex: Regex object for tokenizing
        """
        vocab = set()
        for smi in smiles:
            for token in regex.findall(smi):
                vocab.add(token)

        tokenizer = Tokenizer(
            list(vocab),
            regex,
            begin_token=begin_token,
            end_token=end_token,
            pad_token=pad_token,
            unk_token=unk_token,
        )
        return tokenizer

    @staticmethod
    def from_reactions(
        reaction_data_path: str,
        regex: re.Pattern = REGEX,
        begin_token: str = DEFAULT_BEGIN_TOKEN,
        end_token: str = DEFAULT_END_TOKEN,
        pad_token: str = DEFAULT_PAD_TOKEN,
        unk_token: str = DEFAULT_UNK_TOKEN,
    ):
        """Build the tokenizer from reaction data and a pattern.
        Args:
            reaction_data_path: Path to reaction data
            regex: Regex object for tokenizing
        """
        reaction_dataset = DiskReactionDataset(reaction_data_path, sample_cls=SmilesReactionSample)
        training_dataset = reaction_dataset[DataFold.TRAIN]

        smiles = []
        for sample in training_dataset:
            smiles.append(sample.raw_reactants_smiles)
            smiles.append(sample.raw_products_smiles)

        tokenizer = Tokenizer.from_smiles(
            smiles,
            regex,
            begin_token=begin_token,
            end_token=end_token,
            pad_token=pad_token,
            unk_token=unk_token,
        )
        return tokenizer

    def save_vocab(self, vocab_path: str) -> None:
        """Save the vocabulary to a file."""
        tokens = [key for key, val in sorted(self.vocab_dict.items(), key=lambda x: x[1])]

        tokens_str = ""
        for token in tokens:
            tokens_str += f"{token}\n"

        p = Path(vocab_path)
        p.write_text(tokens_str)

    def __len__(self) -> int:
        return len(self.vocab_dict)

    def encode(
        self,
        smiles: list[str],
        pad: bool = False,
        right_padding: bool = True,
        add_begin_token: bool = False,
        add_end_token: bool = False,
    ) -> tuple[list[list[int]], Optional[list[list[int]]]]:
        """Convert a list of smiles to a list of token ids."""
        tokens_list = [self.regex.findall(smi) for smi in smiles]
        if add_begin_token:
            tokens_list = [[self.begin_token] + tokens for tokens in tokens_list]
        if add_end_token:
            tokens_list = [tokens + [self.end_token] for tokens in tokens_list]

        token_ids_list = self._convert_tokens_to_ids(tokens_list)

        if pad:
            token_ids_list, masks = self.pad_token_ids_list(
                token_ids_list, right_padding=right_padding
            )
            return token_ids_list, masks

        return token_ids_list, None

    def _convert_tokens_to_ids(self, token_data: list[list[str]]) -> list[list[int]]:
        """Convert a list of tokenized sequences to a list of token ids."""
        token_ids_list = []
        for tokens in token_data:
            for token in tokens:
                token_id = self.vocab_dict.get(token)
                if token_id is None:
                    self.met_unk_tokens_dict[token] += 1

            ids = [self.vocab_dict.get(token, self.unk_token_id) for token in tokens]
            token_ids_list.append(ids)
        return token_ids_list

    def pad_token_ids_list(
        self, token_ids_list: list[list[int]], right_padding: bool = True
    ) -> tuple[list[list[int]], list[list[int]]]:
        """Pad sequences to the same length and create masks for padding."""
        pad_length = max([len(token_ids) for token_ids in token_ids_list])
        if right_padding:
            padded_token_ids_list = [
                token_ids + [self.pad_token_id] * (pad_length - len(token_ids))
                for token_ids in token_ids_list
            ]
            masks = [
                [0] * len(token_ids) + [1] * (pad_length - len(token_ids))
                for token_ids in token_ids_list
            ]
        else:
            padded_token_ids_list = [
                [self.pad_token_id] * (pad_length - len(token_ids)) + token_ids
                for token_ids in token_ids_list
            ]
            masks = [
                [1] * (pad_length - len(token_ids)) + [0] * len(token_ids)
                for token_ids in token_ids_list
            ]
        return padded_token_ids_list, masks

    def decode(self, token_ids_list: list[list[int]]) -> list[str]:
        """Convert a list of token ids to a list of sequences."""
        processed_token_ids_list = []
        for token_ids in token_ids_list:
            if token_ids[0] == self.begin_token_id:
                token_ids = token_ids[1:]

            # Remove any tokens after the end token (and end token) if it's there
            if self.end_token_id in token_ids:
                end_token_pos = token_ids.index(self.end_token_id)
                token_ids = token_ids[:end_token_pos]

            processed_token_ids_list.append(token_ids)

        processed_tokens_list = self._convert_ids_to_tokens(processed_token_ids_list)
        smiles = ["".join(tokens) for tokens in processed_tokens_list]
        return smiles

    def _convert_ids_to_tokens(self, token_ids: list[list[int]]) -> list[list[str]]:
        """Convert a list of token ids to a list of tokenized sequences."""
        tokens_list = []
        for ids in token_ids:
            for token_id in ids:
                token = self.decode_vocab_dict.get(token_id)
                if token is None:
                    raise ValueError(f"Token id {token_id} is not recognised")

            tokens = [str(self.decode_vocab_dict.get(token_id)) for token_id in ids]
            tokens_list.append(tokens)
        return tokens_list

    def print_unknown_tokens(self) -> None:
        """Used when training on a new dataset."""
        logger.info(f"{'Token':<10}Count")
        for token, count in self.met_unk_tokens_dict.items():
            logger.info(f"{token:<10}{count}")



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Modified by Microsoft Corporation. See README.md for details.

import itertools
import os.path as osp
from collections import defaultdict
from functools import partial

import numpy as np
import torch
from rdkit import Chem, RDConfig
from rdkit.Chem import AllChem, ChemicalFeatures

# fmt: off

def one_hot_encoding(x, allowable_set, encode_unknown=False):
    """One-hot encoding.

    Parameters
    ----------
    x
        Value to encode.
    allowable_set : list
        The elements of the allowable_set should be of the
        same type as x.
    encode_unknown : bool
        If True, map inputs not in the allowable set to the
        additional last element.

    Returns
    -------
    list
        List of boolean values where at most one value is True.
        The list is of length ``len(allowable_set)`` if ``encode_unknown=False``
        and ``len(allowable_set) + 1`` otherwise.
    """
    if encode_unknown and (allowable_set[-1] is not None):
        allowable_set.append(None)

    if encode_unknown and (x not in allowable_set):
        x = None

    return list(map(lambda s: x == s, allowable_set))

#################################################################
# Atom featurization
#################################################################

def atom_type_one_hot(atom, allowable_set=None, encode_unknown=False):
    """One hot encoding for the type of an atom.

    Parameters
    ----------
    atom : rdkit.Chem.rdchem.Atom
        RDKit atom instance.
    allowable_set : list of str
        Atom types to consider. Default: ``C``, ``N``, ``O``, ``S``, ``F``, ``Si``, ``P``,
        ``Cl``, ``Br``, ``Mg``, ``Na``, ``Ca``, ``Fe``, ``As``, ``Al``, ``I``, ``B``, ``V``,
        ``K``, ``Tl``, ``Yb``, ``Sb``, ``Sn``, ``Ag``, ``Pd``, ``Co``, ``Se``, ``Ti``, ``Zn``,
        ``H``, ``Li``, ``Ge``, ``Cu``, ``Au``, ``Ni``, ``Cd``, ``In``, ``Mn``, ``Zr``, ``Cr``,
        ``Pt``, ``Hg``, ``Pb``.
    encode_unknown : bool
        If True, map inputs not in the allowable set to the
        additional last element. (Default: False)

    Returns
    -------
    list
        List of boolean values where at most one value is True.
    """
    if allowable_set is None:
        allowable_set = ['C', 'N', 'O', 'S', 'F', 'Si', 'P', 'Cl', 'Br', 'Mg', 'Na', 'Ca',
                         'Fe', 'As', 'Al', 'I', 'B', 'V', 'K', 'Tl', 'Yb', 'Sb', 'Sn',
                         'Ag', 'Pd', 'Co', 'Se', 'Ti', 'Zn', 'H', 'Li', 'Ge', 'Cu', 'Au',
                         'Ni', 'Cd', 'In', 'Mn', 'Zr', 'Cr', 'Pt', 'Hg', 'Pb']
    return one_hot_encoding(atom.GetSymbol(), allowable_set, encode_unknown)

def atom_degree_one_hot(atom, allowable_set=None, encode_unknown=False):
    """One hot encoding for the degree of an atom.

    Note that the result will be different depending on whether the Hs are
    explicitly modeled in the graph.

    Parameters
    ----------
    atom : rdkit.Chem.rdchem.Atom
        RDKit atom instance.
    allowable_set : list of int
        Atom degrees to consider. Default: ``0`` - ``10``.
    encode_unknown : bool
        If True, map inputs not in the allowable set to the
        additional last element. (Default: False)

    Returns
    -------
    list
        List of boolean values where at most one value is True.
    """
    if allowable_set is None:
        allowable_set = list(range(11))
    return one_hot_encoding(atom.GetDegree(), allowable_set, encode_unknown)

def atom_implicit_valence_one_hot(atom, allowable_set=None, encode_unknown=False):
    """One hot encoding for the implicit valence of an atom.

    Parameters
    ----------
    atom : rdkit.Chem.rdchem.Atom
        RDKit atom instance.
    allowable_set : list of int
        Atom implicit valences to consider. Default: ``0`` - ``6``.
    encode_unknown : bool
        If True, map inputs not in the allowable set to the
        additional last element. (Default: False)

    Returns
    -------
    list
        List of boolean values where at most one value is True.
    """
    if allowable_set is None:
        allowable_set = list(range(7))
    return one_hot_encoding(atom.GetImplicitValence(), allowable_set, encode_unknown)

# pylint: disable=I1101
def atom_hybridization_one_hot(atom, allowable_set=None, encode_unknown=False):
    """One hot encoding for the hybridization of an atom.

    Parameters
    ----------
    atom : rdkit.Chem.rdchem.Atom
        RDKit atom instance.
    allowable_set : list of rdkit.Chem.rdchem.HybridizationType
        Atom hybridizations to consider. Default: ``Chem.rdchem.HybridizationType.SP``,
        ``Chem.rdchem.HybridizationType.SP2``, ``Chem.rdchem.HybridizationType.SP3``,
        ``Chem.rdchem.HybridizationType.SP3D``, ``Chem.rdchem.HybridizationType.SP3D2``.
    encode_unknown : bool
        If True, map inputs not in the allowable set to the
        additional last element. (Default: False)

    Returns
    -------
    list
        List of boolean values where at most one value is True.
    """
    if allowable_set is None:
        allowable_set = [Chem.rdchem.HybridizationType.SP,
                         Chem.rdchem.HybridizationType.SP2,
                         Chem.rdchem.HybridizationType.SP3,
                         Chem.rdchem.HybridizationType.SP3D,
                         Chem.rdchem.HybridizationType.SP3D2]
    return one_hot_encoding(atom.GetHybridization(), allowable_set, encode_unknown)

def atom_total_num_H_one_hot(atom, allowable_set=None, encode_unknown=False):
    """One hot encoding for the total number of Hs of an atom.

    Parameters
    ----------
    atom : rdkit.Chem.rdchem.Atom
        RDKit atom instance.
    allowable_set : list of int
        Total number of Hs to consider. Default: ``0`` - ``4``.
    encode_unknown : bool
        If True, map inputs not in the allowable set to the
        additional last element. (Default: False)

    Returns
    -------
    list
        List of boolean values where at most one value is True.
    """
    if allowable_set is None:
        allowable_set = list(range(5))
    return one_hot_encoding(atom.GetTotalNumHs(), allowable_set, encode_unknown)

def atom_formal_charge(atom):
    """Get formal charge for an atom.

    Parameters
    ----------
    atom : rdkit.Chem.rdchem.Atom
        RDKit atom instance.

    Returns
    -------
    list
        List containing one int only.
    """
    return [atom.GetFormalCharge()]

def atom_partial_charge(atom):
    """Get Gasteiger partial charge for an atom.

    For using this function, you must have called ``AllChem.ComputeGasteigerCharges(mol)``
    to compute Gasteiger charges.

    Occasionally, we can get nan or infinity Gasteiger charges, in which case we will set
    the result to be 0.

    Parameters
    ----------
    atom : rdkit.Chem.rdchem.Atom
        RDKit atom instance.

    Returns
    -------
    list
        List containing one float only.
    """
    gasteiger_charge = atom.GetProp('_GasteigerCharge')
    if gasteiger_charge in ['-nan', 'nan', '-inf', 'inf']:
        gasteiger_charge = 0
    return [float(gasteiger_charge)]

def atom_num_radical_electrons(atom):
    """Get the number of radical electrons for an atom.

    Parameters
    ----------
    atom : rdkit.Chem.rdchem.Atom
        RDKit atom instance.

    Returns
    -------
    list
        List containing one int only.
    """
    return [atom.GetNumRadicalElectrons()]

def atom_is_aromatic(atom):
    """Get whether the atom is aromatic.

    Parameters
    ----------
    atom : rdkit.Chem.rdchem.Atom
        RDKit atom instance.

    Returns
    -------
    list
        List containing one bool only.
    """
    return [atom.GetIsAromatic()]

def atom_chiral_tag_one_hot(atom, allowable_set=None, encode_unknown=False):
    """One hot encoding for the chiral tag of an atom.

    Parameters
    ----------
    atom : rdkit.Chem.rdchem.Atom
        RDKit atom instance.
    allowable_set : list of rdkit.Chem.rdchem.ChiralType
        Chiral tags to consider. Default: ``rdkit.Chem.rdchem.ChiralType.CHI_UNSPECIFIED``,
        ``rdkit.Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CW``,
        ``rdkit.Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CCW``,
        ``rdkit.Chem.rdchem.ChiralType.CHI_OTHER``.
    encode_unknown : bool
        If True, map inputs not in the allowable set to the
        additional last element. (Default: False)

    Returns
    -------
    list
        List containing one bool only.
    """
    if allowable_set is None:
        allowable_set = [Chem.rdchem.ChiralType.CHI_UNSPECIFIED,
                         Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CW,
                         Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CCW,
                         Chem.rdchem.ChiralType.CHI_OTHER]
    return one_hot_encoding(atom.GetChiralTag(), allowable_set, encode_unknown)

class ConcatFeaturizer(object):
    """Concatenate the evaluation results of multiple functions as a single feature.

    Parameters
    ----------
    func_list : list
        List of functions for computing molecular descriptors from objects of a same
        particular data type, e.g. ``rdkit.Chem.rdchem.Atom``. Each function is of signature
        ``func(data_type) -> list of float or bool or int``. The resulting order of
        the features will follow that of the functions in the list.
    """
    def __init__(self, func_list):
        self.func_list = func_list

    def __call__(self, x):
        """Featurize the input data.

        Parameters
        ----------
        x :
            Data to featurize.

        Returns
        -------
        list
            List of feature values, which can be of type bool, float or int.
        """
        return list(itertools.chain.from_iterable(
            [func(x) for func in self.func_list]))

class BaseAtomFeaturizer(object):
    """An abstract class for atom featurizers.

    Loop over all atoms in a molecule and featurize them with the ``featurizer_funcs``.

    **We assume the resulting DGLGraph will not contain any virtual nodes and a node i in the
    graph corresponds to exactly atom i in the molecule.**

    Parameters
    ----------
    featurizer_funcs : dict
        Mapping feature name to the featurization function.
        Each function is of signature ``func(rdkit.Chem.rdchem.Atom) -> list or 1D numpy array``.
    feat_sizes : dict
        Mapping feature name to the size of the corresponding feature. If None, they will be
        computed when needed. Default: None.
    """
    def __init__(self, featurizer_funcs, feat_sizes=None):
        self.featurizer_funcs = featurizer_funcs
        if feat_sizes is None:
            feat_sizes = dict()
        self._feat_sizes = feat_sizes

    def feat_size(self, feat_name=None):
        """Get the feature size for ``feat_name``.

        When there is only one feature, users do not need to provide ``feat_name``.

        Parameters
        ----------
        feat_name : str
            Feature for query.

        Returns
        -------
        int
            Feature size for the feature with name ``feat_name``. Default to None.
        """
        if feat_name is None:
            assert len(self.featurizer_funcs) == 1, \
                'feat_name should be provided if there are more than one features'
            feat_name = list(self.featurizer_funcs.keys())[0]

        if feat_name not in self.featurizer_funcs:
            return ValueError('Expect feat_name to be in {}, got {}'.format(
                list(self.featurizer_funcs.keys()), feat_name))

        if feat_name not in self._feat_sizes:
            atom = Chem.MolFromSmiles('C').GetAtomWithIdx(0)
            self._feat_sizes[feat_name] = len(self.featurizer_funcs[feat_name](atom))

        return self._feat_sizes[feat_name]

    def __call__(self, mol):
        """Featurize all atoms in a molecule.

        Parameters
        ----------
        mol : rdkit.Chem.rdchem.Mol
            RDKit molecule instance.

        Returns
        -------
        dict
            For each function in self.featurizer_funcs with the key ``k``, store the computed
            feature under the key ``k``. Each feature is a tensor of dtype float32 and shape
            (N, M), where N is the number of atoms in the molecule.
        """
        num_atoms = mol.GetNumAtoms()
        atom_features = defaultdict(list)

        # Compute features for each atom
        for i in range(num_atoms):
            atom = mol.GetAtomWithIdx(i)
            for feat_name, feat_func in self.featurizer_funcs.items():
                atom_features[feat_name].append(feat_func(atom))

        # Stack the features and convert them to float arrays
        processed_features = dict()
        for feat_name, feat_list in atom_features.items():
            feat = np.stack(feat_list)
            processed_features[feat_name] = feat.astype(np.float32)

        return processed_features

class CanonicalAtomFeaturizer(BaseAtomFeaturizer):
    """A default featurizer for atoms.

    The atom features include:

    * **One hot encoding of the atom type**. The supported atom types include
      ``C``, ``N``, ``O``, ``S``, ``F``, ``Si``, ``P``, ``Cl``, ``Br``, ``Mg``,
      ``Na``, ``Ca``, ``Fe``, ``As``, ``Al``, ``I``, ``B``, ``V``, ``K``, ``Tl``,
      ``Yb``, ``Sb``, ``Sn``, ``Ag``, ``Pd``, ``Co``, ``Se``, ``Ti``, ``Zn``,
      ``H``, ``Li``, ``Ge``, ``Cu``, ``Au``, ``Ni``, ``Cd``, ``In``, ``Mn``, ``Zr``,
      ``Cr``, ``Pt``, ``Hg``, ``Pb``.
    * **One hot encoding of the atom degree**. The supported possibilities
      include ``0 - 10``.
    * **One hot encoding of the number of implicit Hs on the atom**. The supported
      possibilities include ``0 - 6``.
    * **Formal charge of the atom**.
    * **Number of radical electrons of the atom**.
    * **One hot encoding of the atom hybridization**. The supported possibilities include
      ``SP``, ``SP2``, ``SP3``, ``SP3D``, ``SP3D2``.
    * **Whether the atom is aromatic**.
    * **One hot encoding of the number of total Hs on the atom**. The supported possibilities
      include ``0 - 4``.

    **We assume the resulting DGLGraph will not contain any virtual nodes.**

    Parameters
    ----------
    atom_data_field : str
        Name for storing atom features in DGLGraphs, default to 'h'.
    """
    def __init__(self, atom_data_field='h'):
        super(CanonicalAtomFeaturizer, self).__init__(
            featurizer_funcs={atom_data_field: ConcatFeaturizer(
                [atom_type_one_hot,
                 atom_degree_one_hot,
                 atom_implicit_valence_one_hot,
                 atom_formal_charge,
                 atom_num_radical_electrons,
                 atom_hybridization_one_hot,
                 atom_is_aromatic,
                 atom_total_num_H_one_hot]
            )})

class WeaveAtomFeaturizer(object):
    """Atom featurizer in Weave.

    The atom featurization performed in `Molecular Graph Convolutions: Moving Beyond Fingerprints
    <https://arxiv.org/abs/1603.00856>`__, which considers:

    * atom types
    * chirality
    * formal charge
    * partial charge
    * aromatic atom
    * hybridization
    * hydrogen bond donor
    * hydrogen bond acceptor
    * the number of rings the atom belongs to for ring size between 3 and 8

    **We assume the resulting DGLGraph will not contain any virtual nodes.**

    Parameters
    ----------
    atom_data_field : str
        Name for storing atom features in DGLGraphs, default to 'h'.
    atom_types : list of str or None
        Atom types to consider for one-hot encoding. If None, we will use a default
        choice of ``'H', 'C', 'N', 'O', 'F', 'P', 'S', 'Cl', 'Br', 'I'``.
    chiral_types : list of Chem.rdchem.ChiralType or None
        Atom chirality to consider for one-hot encoding. If None, we will use a default
        choice of ``Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CW,
        Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CCW``.
    hybridization_types : list of Chem.rdchem.HybridizationType or None
        Atom hybridization types to consider for one-hot encoding. If None, we will use a
        default choice of ``Chem.rdchem.HybridizationType.SP, Chem.rdchem.HybridizationType.SP2,
        Chem.rdchem.HybridizationType.SP3``.
    """
    def __init__(self, atom_data_field='h', atom_types=None, chiral_types=None,
                 hybridization_types=None):
        super(WeaveAtomFeaturizer, self).__init__()

        self._atom_data_field = atom_data_field

        if atom_types is None:
            atom_types = ['H', 'C', 'N', 'O', 'F', 'P', 'S', 'Cl', 'Br', 'I']
        self._atom_types = atom_types

        if chiral_types is None:
            chiral_types = [Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CW,
                            Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CCW]
        self._chiral_types = chiral_types

        if hybridization_types is None:
            hybridization_types = [Chem.rdchem.HybridizationType.SP,
                                   Chem.rdchem.HybridizationType.SP2,
                                   Chem.rdchem.HybridizationType.SP3]
        self._hybridization_types = hybridization_types

        self._featurizer = ConcatFeaturizer([
            partial(atom_type_one_hot, allowable_set=atom_types, encode_unknown=True),
            partial(atom_chiral_tag_one_hot, allowable_set=chiral_types),
            atom_formal_charge, atom_partial_charge, atom_is_aromatic,
            partial(atom_hybridization_one_hot, allowable_set=hybridization_types)
        ])

    def feat_size(self):
        """Get the feature size.

        Returns
        -------
        int
            Feature size.
        """
        mol = Chem.MolFromSmiles('C')
        feats = self(mol)[self._atom_data_field]

        return feats.shape[-1]

    def get_donor_acceptor_info(self, mol_feats):
        """Bookkeep whether an atom is donor/acceptor for hydrogen bonds.

        Parameters
        ----------
        mol_feats : tuple of rdkit.Chem.rdMolChemicalFeatures.MolChemicalFeature
            Features for molecules.

        Returns
        -------
        is_donor : dict
            Mapping atom ids to binary values indicating whether atoms
            are donors for hydrogen bonds
        is_acceptor : dict
            Mapping atom ids to binary values indicating whether atoms
            are acceptors for hydrogen bonds
        """
        is_donor = defaultdict(bool)
        is_acceptor = defaultdict(bool)
        # Get hydrogen bond donor/acceptor information
        for feats in mol_feats:
            if feats.GetFamily() == 'Donor':
                nodes = feats.GetAtomIds()
                for u in nodes:
                    is_donor[u] = True
            elif feats.GetFamily() == 'Acceptor':
                nodes = feats.GetAtomIds()
                for u in nodes:
                    is_acceptor[u] = True

        return is_donor, is_acceptor

    def __call__(self, mol):
        """Featurizes the input molecule.

        Parameters
        ----------
        mol : rdkit.Chem.rdchem.Mol
            RDKit molecule instance.

        Returns
        -------
        dict
            Mapping atom_data_field as specified in the input argument to the atom
            features, which is a float32 tensor of shape (N, M), N is the number of
            atoms and M is the feature size.
        """
        atom_features = []

        AllChem.ComputeGasteigerCharges(mol)
        num_atoms = mol.GetNumAtoms()

        # Get information for donor and acceptor
        fdef_name = osp.join(RDConfig.RDDataDir, 'BaseFeatures.fdef')
        mol_featurizer = ChemicalFeatures.BuildFeatureFactory(fdef_name)
        mol_feats = mol_featurizer.GetFeaturesForMol(mol)
        is_donor, is_acceptor = self.get_donor_acceptor_info(mol_feats)

        # Get a symmetrized smallest set of smallest rings
        # Following the practice from Chainer Chemistry (https://github.com/chainer/
        # chainer-chemistry/blob/da2507b38f903a8ee333e487d422ba6dcec49b05/chainer_chemistry/
        # dataset/preprocessors/weavenet_preprocessor.py)
        sssr = Chem.GetSymmSSSR(mol)

        for i in range(num_atoms):
            atom = mol.GetAtomWithIdx(i)
            # Features that can be computed directly from RDKit atom instances, which is a list
            feats = self._featurizer(atom)
            # Donor/acceptor indicator
            feats.append(float(is_donor[i]))
            feats.append(float(is_acceptor[i]))
            # Count the number of rings the atom belongs to for ring size between 3 and 8
            count = [0 for _ in range(3, 9)]
            for ring in sssr:
                ring_size = len(ring)
                if i in ring and 3 <= ring_size <= 8:
                    count[ring_size - 3] += 1
            feats.extend(count)
            atom_features.append(feats)
        atom_features = np.stack(atom_features)

        return {self._atom_data_field: atom_features.astype(np.float32)}

def bond_type_one_hot(bond, allowable_set=None, encode_unknown=False):
    """One hot encoding for the type of a bond.

    Parameters
    ----------
    bond : rdkit.Chem.rdchem.Bond
        RDKit bond instance.
    allowable_set : list of Chem.rdchem.BondType
        Bond types to consider. Default: ``Chem.rdchem.BondType.SINGLE``,
        ``Chem.rdchem.BondType.DOUBLE``, ``Chem.rdchem.BondType.TRIPLE``,
        ``Chem.rdchem.BondType.AROMATIC``.
    encode_unknown : bool
        If True, map inputs not in the allowable set to the
        additional last element. (Default: False)

    Returns
    -------
    list
        List of boolean values where at most one value is True.
    """
    if allowable_set is None:
        allowable_set = [Chem.rdchem.BondType.SINGLE,
                         Chem.rdchem.BondType.DOUBLE,
                         Chem.rdchem.BondType.TRIPLE,
                         Chem.rdchem.BondType.AROMATIC]
    return one_hot_encoding(bond.GetBondType(), allowable_set, encode_unknown)

def bond_is_conjugated(bond):
    """Get whether the bond is conjugated.

    Parameters
    ----------
    bond : rdkit.Chem.rdchem.Bond
        RDKit bond instance.

    Returns
    -------
    list
        List containing one bool only.
    """
    return [bond.GetIsConjugated()]

def bond_is_in_ring(bond):
    """Get whether the bond is in a ring of any size.

    Parameters
    ----------
    bond : rdkit.Chem.rdchem.Bond
        RDKit bond instance.

    Returns
    -------
    list
        List containing one bool only.
    """
    return [bond.IsInRing()]

def bond_stereo_one_hot(bond, allowable_set=None, encode_unknown=False):
    """One hot encoding for the stereo configuration of a bond.

    Parameters
    ----------
    bond : rdkit.Chem.rdchem.Bond
        RDKit bond instance.
    allowable_set : list of rdkit.Chem.rdchem.BondStereo
        Stereo configurations to consider. Default: ``rdkit.Chem.rdchem.BondStereo.STEREONONE``,
        ``rdkit.Chem.rdchem.BondStereo.STEREOANY``, ``rdkit.Chem.rdchem.BondStereo.STEREOZ``,
        ``rdkit.Chem.rdchem.BondStereo.STEREOE``, ``rdkit.Chem.rdchem.BondStereo.STEREOCIS``,
        ``rdkit.Chem.rdchem.BondStereo.STEREOTRANS``.
    encode_unknown : bool
        If True, map inputs not in the allowable set to the
        additional last element. (Default: False)

    Returns
    -------
    list
        List of boolean values where at most one value is True.
    """
    if allowable_set is None:
        allowable_set = [Chem.rdchem.BondStereo.STEREONONE,
                         Chem.rdchem.BondStereo.STEREOANY,
                         Chem.rdchem.BondStereo.STEREOZ,
                         Chem.rdchem.BondStereo.STEREOE,
                         Chem.rdchem.BondStereo.STEREOCIS,
                         Chem.rdchem.BondStereo.STEREOTRANS]
    return one_hot_encoding(bond.GetStereo(), allowable_set, encode_unknown)

class BaseBondFeaturizer(object):
    """An abstract class for bond featurizers.
    Loop over all bonds in a molecule and featurize them with the ``featurizer_funcs``.
    We assume the constructed ``DGLGraph`` is a bi-directed graph where the **i** th bond in the
    molecule, i.e. ``mol.GetBondWithIdx(i)``, corresponds to the **(2i)**-th and **(2i+1)**-th edges
    in the DGLGraph.

    **We assume the resulting DGLGraph will be created with :func:`smiles_to_bigraph` without
    self loops.**

    Parameters
    ----------
    featurizer_funcs : dict
        Mapping feature name to the featurization function.
        Each function is of signature ``func(rdkit.Chem.rdchem.Bond) -> list or 1D numpy array``.
    feat_sizes : dict
        Mapping feature name to the size of the corresponding feature. If None, they will be
        computed when needed. Default: None.
    self_loop : bool
        Whether self loops will be added. Default to False. If True, it will use an additional
        column of binary values to indicate the identity of self loops in each bond feature.
        The features of the self loops will be zero except for the additional columns.
    """
    def __init__(self, featurizer_funcs, feat_sizes=None, self_loop=False):
        self.featurizer_funcs = featurizer_funcs
        if feat_sizes is None:
            feat_sizes = dict()
        self._feat_sizes = feat_sizes
        self._self_loop = self_loop

    def feat_size(self, feat_name=None):
        """Get the feature size for ``feat_name``.

        When there is only one feature, users do not need to provide ``feat_name``.

        Parameters
        ----------
        feat_name : str
            Feature for query.

        Returns
        -------
        int
            Feature size for the feature with name ``feat_name``. Default to None.
        """
        if feat_name is None:
            assert len(self.featurizer_funcs) == 1, \
                'feat_name should be provided if there are more than one features'
            feat_name = list(self.featurizer_funcs.keys())[0]

        if feat_name not in self.featurizer_funcs:
            return ValueError('Expect feat_name to be in {}, got {}'.format(
                list(self.featurizer_funcs.keys()), feat_name))

        mol = Chem.MolFromSmiles('CCO')
        feats = self(mol)

        return feats[feat_name].shape[1]

    def __call__(self, mol):
        """Featurize all bonds in a molecule.

        Parameters
        ----------
        mol : rdkit.Chem.rdchem.Mol
            RDKit molecule instance.

        Returns
        -------
        dict
            For each function in self.featurizer_funcs with the key ``k``, store the computed
            feature under the key ``k``. Each feature is a tensor of dtype float32 and shape
            (N, M), where N is the number of atoms in the molecule.
        """
        num_bonds = mol.GetNumBonds()
        bond_features = defaultdict(list)

        # Compute features for each bond
        for i in range(num_bonds):
            bond = mol.GetBondWithIdx(i)
            for feat_name, feat_func in self.featurizer_funcs.items():
                feat = feat_func(bond)
                bond_features[feat_name].extend([feat, feat.copy()])

        # Stack the features and convert them to float arrays
        processed_features = dict()
        for feat_name, feat_list in bond_features.items():
            feat = np.stack(feat_list)
            processed_features[feat_name] = feat.astype(np.float32)

        if self._self_loop and num_bonds > 0:
            num_atoms = mol.GetNumAtoms()
            for feat_name in processed_features:
                feats = processed_features[feat_name]
                feats = torch.cat([feats, torch.zeros(feats.shape[0], 1)], dim=1)
                self_loop_feats = torch.zeros(num_atoms, feats.shape[1])
                self_loop_feats[:, -1] = 1
                feats = torch.cat([feats, self_loop_feats], dim=0)
                processed_features[feat_name] = feats

        if self._self_loop and num_bonds == 0:
            num_atoms = mol.GetNumAtoms()
            toy_mol = Chem.MolFromSmiles('CO')
            processed_features = self(toy_mol)
            for feat_name in processed_features:
                feats = processed_features[feat_name]
                feats = torch.zeros(num_atoms, feats.shape[1])
                feats[:, -1] = 1
                processed_features[feat_name] = feats

        return processed_features

class CanonicalBondFeaturizer(BaseBondFeaturizer):
    """A default featurizer for bonds.

    The bond features include:
    * **One hot encoding of the bond type**. The supported bond types include
      ``SINGLE``, ``DOUBLE``, ``TRIPLE``, ``AROMATIC``.
    * **Whether the bond is conjugated.**.
    * **Whether the bond is in a ring of any size.**
    * **One hot encoding of the stereo configuration of a bond**. The supported bond stereo
      configurations include ``STEREONONE``, ``STEREOANY``, ``STEREOZ``, ``STEREOE``,
      ``STEREOCIS``, ``STEREOTRANS``.

    **We assume the resulting DGLGraph will be created with :func:`smiles_to_bigraph` without
    self loops.**

    Parameters
    ----------
    bond_data_field : str
        Name for storing bond features in DGLGraphs, default to ``'e'``.
    self_loop : bool
        Whether self loops will be added. Default to False. If True, it will use an additional
        column of binary values to indicate the identity of self loops. The feature of the
        self loops will be zero except for the additional column.
    """
    def __init__(self, bond_data_field='e', self_loop=False):
        super(CanonicalBondFeaturizer, self).__init__(
            featurizer_funcs={bond_data_field: ConcatFeaturizer(
                [bond_type_one_hot,
                 bond_is_conjugated,
                 bond_is_in_ring,
                 bond_stereo_one_hot]
            )}, self_loop=self_loop)



import random
from typing import Sequence

import numpy as np
from syntheseus import Bag, Molecule, Reaction
from syntheseus.interface.molecule import SMILES_SEPARATOR
from syntheseus.interface.reaction import ReactionMetaData
from syntheseus.reaction_prediction.inference_base import ExternalForwardReactionModel
from syntheseus.reaction_prediction.utils.inference import process_raw_smiles_outputs_forwards

from retrochimera.inference.smiles_transformer import AbstractSmilesTransformerModel


class SmilesTransformerForwardModel(
    AbstractSmilesTransformerModel[Bag[Molecule], Reaction], ExternalForwardReactionModel
):
    def _augment_input(self, input: Bag[Molecule]) -> list[str]:
        from root_aligned.preprocessing.generate_PtoR_data import clear_map_canonical_smiles

        augmented_input = []

        # Obtain `reactant_roots`
        all_reactants_atom_ids = [
            [i for i in range(reactant.rdkit_mol.GetNumAtoms())] for reactant in input
        ]
        max_times = np.prod(
            [len(per_reactant_atom_ids) for per_reactant_atom_ids in all_reactants_atom_ids]
        )
        reactant_roots = [[-1 for _ in input]]

        while len(reactant_roots) < min(self.augmentation_size, max_times):
            roots = [random.choice(all_reactants_atom_ids[k]) for k in range(len(input))]
            if roots not in reactant_roots:
                reactant_roots.append(roots)

        if len(reactant_roots) < self.augmentation_size:
            reactant_roots.extend(
                random.choices(reactant_roots, k=self.augmentation_size - len(reactant_roots))
            )

        for k in range(self.augmentation_size):
            tmp = list(zip(input, reactant_roots[k]))
            reactant_k, reactant_roots_k = [i[0] for i in tmp], [i[1] for i in tmp]

            aligned_reactants = []
            for i, reactant in enumerate(reactant_k):
                reactant_root = reactant_roots_k[i]
                reactant_smi = clear_map_canonical_smiles(
                    reactant.smiles, canonical=True, root=reactant_root
                )
                aligned_reactants.append(reactant_smi)

            augmented_input.append(SMILES_SEPARATOR.join(sorted(aligned_reactants)))

        return augmented_input

    def _process_raw_smiles_outputs(
        self, input: Bag[Molecule], output_list: list[str], metadata_list: list[ReactionMetaData]
    ) -> Sequence[Reaction]:
        return process_raw_smiles_outputs_forwards(input, output_list, metadata_list)



"""Script for evaluating model ensembles using automatically optimized ensembling weights.

In contrast to `eval.py` this script is "offline": it uses pre-generated outputs from the former to
rapidly try many ensembling configurations and hyperparameters.
"""

import glob
import json
import shutil
from collections import defaultdict
from dataclasses import dataclass
from itertools import combinations
from pathlib import Path
from typing import Any, Optional

import numpy as np
import torch
from omegaconf import MISSING
from syntheseus import Bag, Molecule, SingleProductReaction
from syntheseus.reaction_prediction.data.dataset import DataFold, DiskReactionDataset
from syntheseus.reaction_prediction.data.reaction_sample import ReactionSample
from syntheseus.reaction_prediction.utils.config import get_config as cli_get_config
from syntheseus.reaction_prediction.utils.metrics import TopKMetricsAccumulator
from syntheseus.reaction_prediction.utils.misc import set_random_seed

from retrochimera.inference.retrochimera import combine_results
from retrochimera.utils.logging import get_logger
from retrochimera.utils.misc import convert_camel_to_snake

logger = get_logger(__name__)


@dataclass
class EvalEnsemblesConfig:
    data_dir: str = MISSING  # Directory containing the data
    results_dir: str = MISSING  # Directory containing results from evaluating all of the models
    output_dir: str = MISSING  # Directory to use for saving the result
    dataset: str = "uspto"  # Dataset name used to look up "eval_results_{dataset}" subdirectories
    num_repeats: int = 5  # Number of repeats of the weight optimization process to be averaged out

    # Options below are for saving ensemble checkpoints, which are constructed by copying original
    # model checkpoints and saving them in a separate subdirectory together with a `models.json`
    # file containing the ensembling weights. Either both or neither of these paths should be set.
    input_checkpoints_dir: Optional[str] = None  # Directory containing model checkpoints
    output_checkpoints_dir: Optional[str] = None  # Directory for saving ensemble checkpoints
    shorten_model_names: bool = True  # Whether to shorten model names when constructing checkpoints


def make_mol(smiles: str) -> Molecule:
    try:
        return Molecule(smiles, make_rdkit_mol=False)
    except Exception:
        return Molecule(smiles, canonicalize=False, make_rdkit_mol=False)


def load_ground_truths(
    data_dir: str, folds: list[DataFold]
) -> dict[DataFold, list[tuple[Bag[Molecule], Bag[Molecule]]]]:
    """Load ground truth outputs for all reactions in the dataset in order."""
    dataset = DiskReactionDataset(data_dir, sample_cls=ReactionSample)
    return {fold: [(r.products, r.reactants) for r in dataset[fold]] for fold in folds}


def load_predicted_reactants(data: dict[str, Any]) -> list[list[list[Molecule]]]:
    raw_reactants: list[list[list[Molecule]]] = []
    for predictions in data["predictions"]:
        if isinstance(predictions, dict):
            predictions = predictions["predictions"]  # Handle older output format.

        raw_reactants.append(  # Handle older output format.
            [
                [make_mol(r["smiles"]) for r in p.get("reactants", p.get("output"))]
                for p in predictions
            ]
        )

    return raw_reactants


def eval_model(
    results: list[list[Bag[Molecule]]], ground_truths: list[Bag[Molecule]]
) -> list[float]:
    metrics = TopKMetricsAccumulator(max_num_results=100)

    assert len(results) == len(ground_truths)
    for result, ground_truth in zip(results, ground_truths):
        metrics.add([r == ground_truth for r in result])

    return metrics.top_k


def eval_ensemble(
    model_results: list[list[list[Bag[Molecule]]]],
    ground_truths: list[Bag[Molecule]],
    model_weights: list[list[float]],
) -> list[float]:
    metrics = TopKMetricsAccumulator(max_num_results=100)

    dummy_product = Molecule("")  # Product is not important here.
    for results, ground_truth in zip(zip(*model_results), ground_truths):
        results_combined = combine_results(
            input=dummy_product,
            model_results=[
                [SingleProductReaction(product=dummy_product, reactants=r) for r in result]
                for result in results
            ],
            model_weights=model_weights,
            num_results=100,
        )

        metrics.add([r.reactants == ground_truth for r in results_combined])

    return metrics.top_k


def get_weights_from_params(params: torch.Tensor) -> torch.Tensor:
    weights = torch.exp(params)  # Transform into non-negative.
    weights = torch.cumsum(weights, dim=1)  # Transform into increasing.
    weights = torch.cumsum(weights, dim=1)  # Transform into convex.
    weights = weights.flip(dims=[1])
    return weights / weights.mean()


def optimize_ensemble_weights(
    model_results: list[list[list[Bag[Molecule]]]],
    ground_truths: list[Bag[Molecule]],
    num_results: int = 100,
    initial_learning_rate: float = 0.1,
    initial_temperature: float = 0.1,
    num_steps: int = 1000,
    decay_every_n_steps: int = 25,
    decay_factor: float = 0.9,
    margin: float = 1e-4,
    regularization_loss_weight: float = 0.2,
) -> torch.Tensor:
    num_models = len(model_results)

    rank_list_pairs: list[torch.Tensor] = []
    for results, ground_truth in zip(zip(*model_results), ground_truths):
        result_to_ranks: dict[Bag[Molecule], list[int]] = defaultdict(
            lambda: [num_results] * num_models
        )

        for model_idx, results in enumerate(results):
            for rank, r in enumerate(results):
                result_to_ranks[r][model_idx] = rank

        if ground_truth not in result_to_ranks:
            # If the ground-truth was not recovered at all then there's nothing to optimize.
            continue

        ranks_ground_truth = result_to_ranks[ground_truth]
        for r, ranks in result_to_ranks.items():
            if (
                r == ground_truth
                or all([pi <= ni for (pi, ni) in zip(ranks_ground_truth, ranks)])
                or all([pi >= ni for (pi, ni) in zip(ranks_ground_truth, ranks)])
            ):
                # If one of the results has better rank from all models than another result, then
                # the ensemble will always rank them in the same way; nothing to optimize.
                continue

            # Stack the ranks into a tensor and shift to make them valid indices into an unraveled
            # ensembling weight matrix.
            rank_list_pairs.append(
                torch.as_tensor([ranks_ground_truth, ranks])
                + (num_results + 1) * torch.arange(num_models)
            )

    rank_list_pairs_stacked = torch.stack(rank_list_pairs, dim=0)
    params = torch.nn.parameter.Parameter(data=torch.zeros((num_models, num_results)))

    def compute_loss(temperature: float) -> torch.Tensor:
        weights = get_weights_from_params(params)
        all_weight_ratios = [
            weights[model_idx_1] / weights[model_idx_2]
            for model_idx_1 in range(num_models)
            for model_idx_2 in range(num_models)
            if model_idx_1 != model_idx_2
        ]

        weights = torch.cat([weights, torch.zeros(num_models, 1)], dim=-1).ravel()

        # Map ranks to scores, and sum over the models.
        scores = weights[rank_list_pairs_stacked].sum(dim=-1)

        # Subtract the positive scores (first column) from the negative ones (second column).
        scores[:, 0] *= -1
        scores = scores.sum(dim=-1)

        # Compute the ranking loss.
        loss_ranking = torch.sigmoid((scores + margin) / temperature).sum() / len(ground_truths)

        # Compute the regularization loss.
        loss_regularization = 0.0
        for weight_ratios in all_weight_ratios:
            loss_regularization += (weight_ratios[:-1] - weight_ratios[1:]).abs().mean()

        loss_regularization = loss_regularization / len(all_weight_ratios)
        return loss_ranking + regularization_loss_weight * loss_regularization

    optimizer = torch.optim.Adam([params], lr=initial_learning_rate)
    temperature = initial_temperature

    for step in range(num_steps):
        optimizer.zero_grad()
        loss = compute_loss(temperature)
        loss.backward()
        optimizer.step()

        if (step + 1) % decay_every_n_steps == 0 and step != num_steps - 1:
            temperature *= decay_factor
            for g in optimizer.param_groups:
                g["lr"] *= decay_factor

    return get_weights_from_params(params).detach()


def merge_top_k(
    top_k_1: list[float], num_samples_1: int, top_k_2: list[float], num_samples_2: int
) -> list[float]:
    # Go back from accuracies to number of correct answers (which can be mapped to an integer).
    [num_correct_1, num_correct_2] = [
        [round(v * num_samples) for v in top_k]
        for (top_k, num_samples) in [(top_k_1, num_samples_1), (top_k_2, num_samples_2)]
    ]

    num_correct_total = [v_1 + v_2 for (v_1, v_2) in zip(num_correct_1, num_correct_2)]
    num_samples_total = num_samples_1 + num_samples_2

    return [v / num_samples_total for v in num_correct_total]


def run_from_config(config: EvalEnsemblesConfig) -> None:
    set_random_seed(0)

    if (config.input_checkpoints_dir is None) != (config.output_checkpoints_dir is None):
        raise ValueError(
            "Both or neither of `input_checkpoints_dir` and `output_checkpoints_dir` should be set"
        )

    output_dir = Path(config.output_dir)

    logger.info(f"Loading dataset ground truths from {config.data_dir}")
    ground_truths = {
        fold: [reactants for _, reactants in samples]
        for fold, samples in load_ground_truths(
            data_dir=config.data_dir, folds=[DataFold.VALIDATION, DataFold.TEST]
        ).items()
    }

    # If the results for some of the models are split into subdirectories, make sure these are
    # processed in correct order (we assume a naming scheme of the form `...1`, `...2`, etc).
    result_paths = sorted(
        glob.glob(f"{config.results_dir}/**/eval_results_{config.dataset}/*.json", recursive=True)
    )

    result_paths_joined = "\n".join(result_paths)
    logger.info(f"Found {len(result_paths)} results files:\n{result_paths_joined}")

    model_to_fold_to_results: dict[str, dict[DataFold, list[list[Bag[Molecule]]]]] = defaultdict(
        lambda: defaultdict(list)
    )
    model_to_fold_to_top_k: dict[str, dict[DataFold, list[float]]] = defaultdict(dict)
    model_to_short_name: dict[str, str] = {}
    model_to_model_class: dict[str, str] = {}
    fold_name_to_fold = {f.name.lower(): f for f in DataFold}

    for path in result_paths:
        path_parts = path.split("/")

        [(fold, model_class)] = [
            (fold_name_to_fold[fold_name], model_class)
            for (fold_name, model_class) in zip(path_parts, path_parts[1:])
            if fold_name in fold_name_to_fold
        ]

        if fold == DataFold.TRAIN:
            continue

        logger.info(f"Loading data for model {model_class} and fold {fold}")
        with open(path, "rt") as f:
            data = json.load(f)

            if not data["predictions"]:
                continue

            args = data["eval_args"]

            # Convert case (e.g. "TemplateLocalizationModel" -> "template_localization").
            short_name = convert_camel_to_snake(args["model_class"].removesuffix("Model"))

            # Make sure the model class and fold that we parsed out of the path matches the file.
            assert model_class.startswith(short_name)

            model_to_short_name[model_class] = short_name
            model_to_model_class[model_class] = args["model_class"]

            # Results for Transformer have an incorrect fold recorded in the output files.
            if "smiles_transformer" not in model_class:
                assert args["fold"] == fold.name

            top_k = [float(value) for value in data["top_k"]]

            if fold not in model_to_fold_to_top_k[model_class]:
                model_to_fold_to_top_k[model_class][fold] = top_k
            else:
                model_to_fold_to_top_k[model_class][fold] = merge_top_k(
                    top_k_1=model_to_fold_to_top_k[model_class][fold],
                    num_samples_1=len(model_to_fold_to_results[model_class][fold]),
                    top_k_2=top_k,
                    num_samples_2=len(data["predictions"]),
                )

            for reactants in load_predicted_reactants(data):
                model_to_fold_to_results[model_class][fold].append([Bag(r) for r in reactants])

    models_to_ensemble: list[str] = []
    for model, model_results in model_to_fold_to_results.items():
        if set(model_results.keys()) < {DataFold.VALIDATION, DataFold.TEST}:
            logger.warning(
                f"Model {model} will not be used for ensembling as not all folds were found"
            )
        else:
            assert set(model_results.keys()) == {DataFold.VALIDATION, DataFold.TEST}

            for fold in model_results:
                assert len(ground_truths[fold]) == len(model_results[fold])

            models_to_ensemble.append(model)

    if config.input_checkpoints_dir is not None:
        for model in models_to_ensemble:
            model_ckpt_dir = Path(config.input_checkpoints_dir) / model
            if not model_ckpt_dir.is_dir():
                raise ValueError(
                    f"Ensemble checkpoint saving is on, but {model_ckpt_dir} does not exist"
                )

        if config.shorten_model_names and len(models_to_ensemble) != len(
            set(model_to_short_name[m] for m in models_to_ensemble)
        ):
            raise ValueError(
                "Ensemble checkpoint saving with model name shortening is on, but the model names "
                "are not unique when shortened (i.e. two checkpoints of the same underlying model)."
            )

        if len(models_to_ensemble) > 2:
            logger.warning(
                "Ensemble checkpoint saving is on, so we will save a checkpoint for each "
                f"combination. As {len(models_to_ensemble)} models are being ensembled, this may "
                "use a lot of disk space (we copy all checkpoints involved in each ensemble)."
            )

    model_to_test_top_k = {
        model: model_to_fold_to_top_k[model][DataFold.TEST] for model in model_to_fold_to_top_k
    }

    with open(output_dir / "individual.json", "wt") as f:
        json.dump(model_to_test_top_k, f)

    logger.info("Verifying results for models tested in isolation")
    for model in model_to_fold_to_results:
        for fold in model_to_fold_to_results[model]:
            individual_results = eval_model(
                results=model_to_fold_to_results[model][fold],
                ground_truths=ground_truths[fold],
            )

            # Verify that we get the same accuracies as what was reported.
            assert np.allclose(individual_results, model_to_fold_to_top_k[model][fold])

    models_public = []
    models_private = []
    for model in models_to_ensemble:
        if "smiles_transformer" in model or "template_localization" in model:
            models_private.append(model)
        else:
            models_public.append(model)

    # First, ensemble all model pairs.
    model_subsets_to_ensemble = [list(t) for t in combinations(models_public + models_private, 2)]

    if config.dataset == "uspto":
        # On USPTO-50K RetroKNN's adapter was trained on validation set, inflating the model's
        # perceived performance, and making the ensembling algorithm put excessive weight on
        # RetroKNN. We thus exclude it from the set for the larger ensembles.
        models_public = [model for model in models_public if model != "retro_knn"]

    # Second, also add ensembles containing all public models and a subset of private models.
    for num_private in range(len(models_private) + 1):
        if len(models_public) + num_private < 2:
            continue

        for t in combinations(models_private, num_private):
            new_subset = models_public + list(t)
            if new_subset not in model_subsets_to_ensemble:
                model_subsets_to_ensemble.append(new_subset)

    model_subsets_to_ensemble = [sorted(model_subset) for model_subset in model_subsets_to_ensemble]

    model_subsets_joined = "\n".join([" ".join(ms) for ms in model_subsets_to_ensemble])
    logger.info(f"Ensembling the following subsets of models:\n{model_subsets_joined}")

    all_results: dict[str, dict[str, list]] = defaultdict(dict)
    for model_subset in model_subsets_to_ensemble:
        key = " ".join(model_subset)
        logger.info(f"Running ensembling for {key}")

        all_weights = [
            optimize_ensemble_weights(
                model_results=[
                    model_to_fold_to_results[model][DataFold.VALIDATION] for model in model_subset
                ],
                ground_truths=ground_truths[DataFold.VALIDATION],
            )
            for _ in range(config.num_repeats)
        ]

        results: dict[str, Any] = {}
        results["weights"] = (sum(all_weights) / len(all_weights)).numpy().tolist()

        for fold in [DataFold.VALIDATION, DataFold.TEST]:
            results[f"{fold.name.lower()}_accuracies"] = eval_ensemble(
                model_results=[model_to_fold_to_results[model][fold] for model in model_subset],
                ground_truths=ground_truths[fold],
                model_weights=results["weights"],
            )

        # Dump under a string key for JSON saving to work.
        all_results[key] = results

        if config.output_checkpoints_dir is not None:
            assert config.input_checkpoints_dir is not None

            checkpoint_dir = (
                Path(config.output_checkpoints_dir) / f"ensemble_{'_'.join(model_subset)}"
            )
            checkpoint_dir.mkdir(parents=True, exist_ok=True)

            logger.info(f"Saving ensemble checkpoint to {checkpoint_dir}")

            model_config: dict[str, tuple[str, list[float]]] = {}
            for model in model_subset:
                short_name = model_to_short_name[model]
                subdir_name = short_name if config.shorten_model_names else model

                input_checkpoint_dir = Path(config.input_checkpoints_dir) / model
                output_checkpoint_dir = checkpoint_dir / subdir_name
                output_checkpoint_dir.mkdir(parents=True, exist_ok=True)

                for item in input_checkpoint_dir.iterdir():
                    if item.is_file():
                        shutil.copy(item, output_checkpoint_dir)

                model_config[short_name] = (
                    model_to_model_class[model],
                    results["weights"][model_subset.index(model)],
                )

            with open(checkpoint_dir / "models.json", "wt") as f:
                json.dump(model_config, f)

    with open(output_dir / "ensembles_optimized.json", "wt") as f:
        json.dump(all_results, f)


def main(argv: Optional[list[str]]) -> None:
    config: EvalEnsemblesConfig = cli_get_config(argv=argv, config_cls=EvalEnsemblesConfig)
    run_from_config(config)


if __name__ == "__main__":
    main(argv=None)



import logging
import sys

import tqdm


class TqdmLoggingHandler(logging.StreamHandler):
    def emit(self, record):
        try:
            msg = self.format(record)
            tqdm.tqdm.write(msg)
            self.flush()
        except Exception:
            self.handleError(record)


# Idea borrowed from
# https://github.com/microsoft/DeepSpeed/blob/master/deepspeed/utils/logging.py
def get_logger(name=None, level=logging.INFO) -> logging.Logger:
    """Returns a logger that is configured as:
    - by default INFO level or higher messages are logged out in STDOUT.
    - format includes file name, line number, etc.
    """
    logger = logging.getLogger(name)
    logger.setLevel(level)
    logger.propagate = False

    if logger.hasHandlers():
        # Remove existing handlers so that capsys can capture
        # the output from patched sys.stdout
        for handler in logger.handlers:
            logger.removeHandler(handler)

    log_formatter = logging.Formatter(
        "[%(asctime)s] [%(levelname)s] [%(filename)s:%(lineno)d:%(funcName)s] %(message)s"
    )

    # Send everything to stdout
    handler_out = TqdmLoggingHandler(sys.stdout)
    handler_out.setFormatter(log_formatter)
    logger.addHandler(handler_out)

    return logger


# Delay evaluation of "logger" attribute so that capsys can capture the
# output of this logger.
def __getattr__(name):
    if name == "logger":
        return get_logger(name="feynman", level=logging.INFO)
    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")



import tempfile
from typing import Sequence

import pytest
from rdkit import Chem
from syntheseus import Bag, Molecule, SingleProductReaction
from syntheseus.reaction_prediction.chem.utils import molecule_bag_from_smiles

from retrochimera.chem.fixing import reverse_rxn
from retrochimera.chem.rules import RuleBase, RuleBasedRetrosynthesizer, RulePrediction


def check_result_against_target(result: Sequence[SingleProductReaction], target: str) -> None:
    predictions = {prediction.reactants for prediction in result}
    assert molecule_bag_from_smiles(target) in predictions


def check_results_against_targets(
    results: list[Sequence[SingleProductReaction]], targets: list[str]
) -> None:
    assert len(results) == len(targets)

    for result, target in zip(results, targets):
        check_result_against_target(result, target)


def test_check_results() -> None:
    # Verify that `check_result(s)_against_target(s)` functions trigger the asserts correctly.
    input = Molecule("CC")
    result_1 = [
        SingleProductReaction(product=input, reactants=Bag([Molecule(smiles)]))
        for smiles in ["c1ccccc1", "c1ccccc1N"]
    ]
    result_2 = [
        SingleProductReaction(product=input, reactants=Bag([Molecule(smiles)]))
        for smiles in ["c1ccccc1N", "c1ccccc1O"]
    ]
    target = "c1ccccc1O"

    # `result_1` does not have the correct prediction:
    with pytest.raises(AssertionError):
        check_result_against_target(result=result_1, target=target)

    # Number of results and targets do not match:
    with pytest.raises(AssertionError):
        check_results_against_targets(results=[result_2, result_2], targets=[target])

    # `result_2` has the correct prediction:
    check_result_against_target(result=result_2, target=target)

    # ...but if we check both results, then things fail:
    with pytest.raises(AssertionError):
        check_results_against_targets(results=[result_2, result_1], targets=[target, target])


def make_retrosynthesizer_from_rulebase(rulebase: RuleBase) -> RuleBasedRetrosynthesizer:
    with tempfile.TemporaryDirectory() as temp_dir:
        rulebase.save_to_file(dir=temp_dir)
        model = RuleBasedRetrosynthesizer()
        model.start_server(rulebase_dir=temp_dir)

    return model


def test_synth() -> None:
    sma = """([#6A&h0X3v4+0:20][#8A&h1X2v2+0:38].[#7A&h2X3v3+0:22])>>([#6A&h0X3v4+0:20][#7A&h1X3v3+0:22])
            ([#8A&h0X2v2+0:1][#6A&h3X4v4+0:16])>>([#8A&h1X2v2+0:1])
            ([#6A&h0X3v4+0:5][#8A&h1X2v2+0:39].[#7A&h1X3v3+0:7])>>([#6A&h0X3v4+0:5][#7A&h0X3v3+0:7])
            ([#7A&h0X3v4+:21](=[#8A&h0X&v2+0:25])[#8A&h0X&v-:26])>>([#7A&h2X3v3+0:21])
            ([#7A&h2X3v3+0:2].[#6ah0X3v4+0:3][ClA&h0X&v+0:10])>>([#7A&h1X3v3+0:2][#6ah0X3v4+0:3])
            ([#7A&h2X3v3+0:3].[#6A&h0X3v4+0:2][ClA&h0X&v+0:13])>>([#6A&h0X3v4+0:2][#7A&h1X3v3+0:3])
            ([#8A&h0X2v2+0:15][#6A&h2X4v4+0:17][#6A&h3X4v4+0:16])>>([#8A&h1X2v2+0:15])
            ([#6ah0X3v4+0:19][BrA&h0X&v+0:34].[#6ah0X3v4+0:20][#5A&h0X3v3+0:36]([#8A&h1X2v2+0:35])[#8A&h1X2v2+0:37])>>([#6ah0X3v4+0:19]-[#6ah0X3v4+0:20])
            ([#7A&h0X3v3+0:6][#6A&h0X3v4+0:14]([#8A&h0X2v2+0:13][#6A&h0X4v4+0:10]([#6A&h3X4v4+0:9])([#6A&h3X4v4+0:11])[#6A&h3X4v4+0:12])=[#8A&h0X&v2+0:15])>>([#7A&h1X3v3+0:6])
            ([#6A&h2X4v4+0:11][BrA&h0X&v+0:19].[#8A&h1X2v2+0:10])>>([#8A&h0X2v2+0:10][#6A&h2X4v4+0:11])
            """.split()

    rule_base = RuleBase()

    for i in sma:
        rule_base.add_rule(reverse_rxn(i))

    model = make_retrosynthesizer_from_rulebase(rule_base)

    input_smiles = [
        "O=C(Nc1cccc2cnccc12)c1cc([N+](=O)[O-])c(Sc2c(Cl)cncc2Cl)s1",
        "CC(C)n1nc(-c2ccc(O)cc2)c2cccc(Cl)c21",
        "CN(C)C(=O)[C@@H]1CCCN1C(=O)OCc1ccccc1",
        "Cc1ccc(C(C)(C)C)cc1N",
        "CC(C)(C)OC(=O)N1CCN(c2nc(-c3ccnc(NC4CCOCC4)c3)cc3cnccc23)CC1",
        "Cc1ccccc1C(=O)Nc1ccc(C(=O)N2Cc3ccccc3Sc3ncccc32)cc1",
        "O=C(O)CCc1c(/C=C2\\C(=O)Nc3ccccc32)[nH]c2c1C(=O)CCC2",
        "Cn1cnc(-c2cc(C#N)ccn2)c1-c1ccc(-n2cccn2)cc1",
        "C[C@H]1CNCCN1C1CCc2ccc(C(F)(F)F)cc21",
        "CCOP(=O)(COCc1nc2c(N)ncnc2n1CCc1ccccc1)OCC",
    ]
    input_mols = [Molecule(smiles) for smiles in input_smiles]
    results = model.predict(input_mols)

    assert len(results) == len(input_mols)

    # Check that the `input` is set correctly.
    for result, input_mol in zip(results, input_mols):
        for prediction in result:
            assert prediction.product == input_mol


def test_rediscovery() -> None:
    data = [
        ("CCN.O=C(O)c1cccs1>>CCNC(=O)c1cccs1", 645),
        ("CCCCCCN.NCCC(=O)O>>CCCCCCNC(=O)CCN", 645),
        ("CN.O=C(O)c1ccccc1I>>CNC(=O)c1ccccc1I", 645),
        ("Nc1ccccc1.O=C(O)CCl>>O=C(CCl)Nc1ccccc1", 645),
        ("CN.O=C(O)c1cc[nH]c1>>CNC(=O)c1cc[nH]c1", 645),
        ("N#CCC(=O)O.Nc1ccccc1>>N#CCC(=O)Nc1ccccc1", 645),
        ("CC(C)(C)N.O=C(O)CCBr>>CC(C)(C)NC(=O)CCBr", 645),
        ("COCCCCC(=O)OC>>COCCCCC(=O)O", 984),
        ("C=C(C)C(=O)OC>>C=C(C)C(=O)O", 984),
        ("COc1cccc(Cl)c1N>>Nc1c(O)cccc1Cl", 984),
        ("COc1ccc(S)cc1Cl>>Oc1ccc(S)cc1Cl", 984),
        ("COc1cnc(Cl)nc1C>>Cc1nc(Cl)ncc1O", 984),
        ("COc1ccc2ccsc2c1>>Oc1ccc2ccsc2c1", 984),
        ("COC(=O)Cc1ccccc1>>O=C(O)Cc1ccccc1", 984),
        ("O=[N+]([O-])c1ccn[nH]1>>Nc1ccn[nH]1", 319),
        ("COc1ncccc1[N+](=O)[O-]>>COc1ncccc1N", 319),
        ("CNc1ccccc1[N+](=O)[O-]>>CNc1ccccc1N", 319),
        ("O=Cc1ccccc1[N+](=O)[O-]>>Nc1ccccc1C=O", 319),
        ("CCCn1ccc([N+](=O)[O-])n1>>CCCn1ccc(N)n1", 319),
        ("Nc1ccc(F)cc1[N+](=O)[O-]>>Nc1ccc(F)cc1N", 319),
        ("O=[N+]([O-])c1ccn(CCO)n1>>Nc1ccn(CCO)n1", 319),
        ("CCCCBr.OCCCO>>CCCCOCCCO", 630),
        ("C#CCBr.COCCOCCO>>C#CCOCCOCCOC", 630),
        ("C#CCBr.OCC1CCCC1>>C#CCOCC1CCCC1", 630),
        ("CCCCBr.OCc1ccsc1>>CCCCOCc1ccsc1", 630),
        ("CCCCCCO.O=C(O)CBr>>CCCCCCOCC(=O)O", 630),
        ("O=C(O)CBr.OC1CCC1>>O=C(O)COC1CCC1", 630),
        ("BrCCCCBr.OC1CCCC1>>BrCCCCOC1CCCC1", 630),
    ]

    rules = {
        645: "([#6A&h0X3v4+0:20]-[#7A&h1X3v3+0:22])>>([#6A&h0X3v4+0:20]-[#8A&h1X2v2+0:38].[#7A&h2X3v3+0:22])",
        984: "([#8A&h1X2v2+0:1])>>([#8A&h0X2v2+0:1]-[#6A&h3X4v4+0:16])",
        319: "([#7A&h2X3v3+0:21])>>([#7A&h0X3v4+:21](=[#8A&h0X&v2+0:25])-[#8A&h0X&v-:26])",
        630: "([#8A&h0X2v2+0:10]-[#6A&h2X4v4+0:11])>>([#6A&h2X4v4+0:11]-[BrA&h0X&v+0:19].[#8A&h1X2v2+0:10])",
    }

    rule_base = RuleBase()

    for k, v in rules.items():
        rule_base.add_rule(v, rule_id=k)

    model = make_retrosynthesizer_from_rulebase(rule_base)

    input_smiles = [i[0].split(">>")[1] for i in data]
    targets = [Chem.CanonSmiles(i[0].split(">>")[0]) for i in data]

    results = model.predict([Molecule(smiles) for smiles in input_smiles])
    check_results_against_targets(results, targets)


def test_RuleBasedRetrosynthesizer() -> None:
    smiles = "COCCCCC(=O)O"
    target = "COCCCCC(=O)OC"

    rules = {
        645: "([#6A&h0X3v4+0:20]-[#7A&h1X3v3+0:22])>>([#6A&h0X3v4+0:20]-[#8A&h1X2v2+0:38].[#7A&h2X3v3+0:22])",
        984: "([#8A&h1X2v2+0:1])>>([#8A&h0X2v2+0:1]-[#6A&h3X4v4+0:16])",
        319: "([#7A&h2X3v3+0:21])>>([#7A&h0X3v4+:21](=[#8A&h0X&v2+0:25])-[#8A&h0X&v-:26])",
        630: "([#8A&h0X2v2+0:10]-[#6A&h2X4v4+0:11])>>([#6A&h2X4v4+0:11]-[BrA&h0X&v+0:19].[#8A&h1X2v2+0:10])",
    }

    rule_base = RuleBase()

    for k, v in rules.items():
        rule_base.add_rule(v, rule_id=k)

    model = make_retrosynthesizer_from_rulebase(rule_base)

    [result] = model.apply_top_rules(
        [Molecule(smiles, identifier=1)], batch_rules_to_apply=[[RulePrediction(id=984, prob=1.0)]]
    )

    rules_to_apply = model._predict_ranked_rules([Molecule(smiles, identifier=1)] * 7, top_k=4)
    assert len(rules_to_apply) == 7

    check_result_against_target(result, target)


def test_rdchiral_templates() -> None:
    # Here we check RDCHIRAL templates, taken from https://www.science.org/doi/10.1126/sciadv.abe4166

    data = [
        ("c1ccccc1[N+](=O)[O-]>>Nc1ccccc1"),
        ("CS(Cl)(=O)=O.OC1CCCC1>>CS(=O)(=O)OC1CCCC1"),
        ("OC(=O)C1=CC=CC=C1.C1CCNCC1>>O=C(N1CCCCC1)C1=CC=CC=C1"),
        ("CN1CCCC1C(=O)OC(C)(C)C>>CN1CCCC1C(O)=O"),
        ("ClC1=CC=CC=N1.FC1CCNC1>>FC1CCN(C1)C1=CC=CC=N1"),
    ]

    rules = {
        1: "([C:3]-[C:2](=[O;D1;H0:4])-[OH;D1;+0:1])>>(C-[O;H0;D2;+0:1]-[C:2](-[C:3])=[O;D1;H0:4])",
        2: "([NH2;D1;+0:1]-[c:2])>>(O=[N+;H0;D3:1](-[O-])-[c:2])",
        3: "([C:5]-[O;H0;D2;+0:6]-[S;H0;D4;+0:1](-[C;D1;H3:2])(=[O;D1;H0:3])=[O;D1;H0:4])>>(Cl-[S;H0;D4;+0:1](-[C;D1;H3:2])(=[O;D1;H0:3])=[O;D1;H0:4].[C:5]-[OH;D1;+0:6])",
        4: "([C:4]-[N;H0;D3;+0:5](-[C:6])-[C;H0;D3;+0:1](=[O;D1;H0:2])-[c:3])>>(O-[C;H0;D3;+0:1](=[O;D1;H0:2])-[c:3].[C:4]-[NH;D2;+0:5]-[C:6])",
        5: "([C:3]-[C:2](=[O;D1;H0:4])-[OH;D1;+0:1])>>(C-C(-C)(-C)-[O;H0;D2;+0:1]-[C:2](-[C:3])=[O;D1;H0:4])",
        6: "([#7;a:2]:[c;H0;D3;+0:1](:[c:3])-[N;H0;D3;+0:5](-[C:4])-[C:6])>>(Cl-[c;H0;D3;+0:1](:[#7;a:2]):[c:3].[C:4]-[NH;D2;+0:5]-[C:6])",
    }

    rule_base = RuleBase()

    for k, v in rules.items():
        rule_base.add_rule(v, rule_id=k)

    model = make_retrosynthesizer_from_rulebase(rule_base)

    input_smiles = [i.split(">>")[1] for i in data]
    targets = [Chem.CanonSmiles(i.split(">>")[0]) for i in data]

    results = model.predict([Molecule(smiles) for smiles in input_smiles])
    check_results_against_targets(results, targets)



import torch
from torch.optim.lr_scheduler import LambdaLR


def count_occurrences(indices: torch.Tensor) -> torch.Tensor:
    """Convert a tensor of indices into counts of how many each value occurs.

    >>> count_occurrences(torch.as_tensor([0, 1, 1, 3]))
    tensor([1, 2, 0, 1])
    """
    out = torch.zeros(indices.max() + 1, device=indices.device, dtype=indices.dtype)
    return torch.scatter_add(out, 0, indices, torch.ones_like(indices))


def split_select(source: torch.Tensor, indices: torch.Tensor) -> list[torch.Tensor]:
    """Split a given tensor into chunks denoted by a separate set of indices.

    >>> split_select(torch.arange(6), torch.as_tensor([0, 0, 1, 2, 2, 2]))
    (tensor([0, 1]), tensor([2]), tensor([3, 4, 5]))
    """
    return torch.split(source, tuple(count_occurrences(indices).tolist()))


def tensor_to_list(t: torch.Tensor):
    return t.cpu().detach().numpy().tolist()


def get_sorted_ids_and_probs(logits: torch.Tensor, k: int) -> list[tuple[list[int], list[float]]]:
    """Given raw batched logits return `k` most likely indices with their probabilities."""
    # Compute probabilities and extract top indices.
    batch_rule_probs = torch.nn.functional.softmax(logits, dim=1)
    sorted_batch_rule_ids = batch_rule_probs.argsort(dim=1, descending=True)

    # Limit to top rules before moving to CPU and converting to `RulePrediction` objects.
    sorted_batch_rule_ids = sorted_batch_rule_ids[:, :k]
    sorted_batch_rule_probs = batch_rule_probs.gather(index=sorted_batch_rule_ids, dim=1)

    sorted_batch_rule_ids = tensor_to_list(sorted_batch_rule_ids)
    sorted_batch_rule_probs = tensor_to_list(sorted_batch_rule_probs)

    return list(zip(sorted_batch_rule_ids, sorted_batch_rule_probs))


class FuncLR(LambdaLR):
    def get_lr(self):
        return [lmbda(self.last_epoch) for lmbda in self.lr_lambdas]



from typing import Any, Sequence

from syntheseus import Molecule, SingleProductReaction
from syntheseus.reaction_prediction.inference_base import ExternalBackwardReactionModel
from syntheseus.reaction_prediction.utils.inference import get_unique_file_in_dir

from retrochimera.chem.rules import RuleBasedRetrosynthesizer, RulePrediction


class TemplateClassificationModel(RuleBasedRetrosynthesizer, ExternalBackwardReactionModel):
    """Wrapper for a model performing template classification backed by a given molecule encoder."""

    def __init__(self, *args, **kwargs) -> None:
        """Initializes the TemplateClassification model wrapper.

        Assumed format of the model directory:
        - `model_dir` contains the model checkpoint as the only `*.ckpt` file
        """

        from retrochimera.models.template_classification import MCCModel

        super().__init__(*args, **kwargs)
        self.start_server(rulebase_dir=self.model_dir)

        self.model = MCCModel.load_from_checkpoint(
            get_unique_file_in_dir(self.model_dir, pattern="*.ckpt")
        )
        self.model.to(self.device)
        self.model.eval()

    def get_model_info(self) -> dict[str, Any]:
        return self.model.hparams

    def get_parameters(self):
        return self.model.parameters()

    def _predict_ranked_rules(
        self, targets: list[Molecule], top_k: int = 50
    ) -> list[list[RulePrediction]]:
        from retrochimera.utils.pytorch import get_sorted_ids_and_probs

        mol_outputs = self.model.encoder.forward_raw(targets, device=self.device).mol_outputs
        assert mol_outputs is not None

        return [
            [RulePrediction(id=id, prob=prob) for id, prob in zip(rule_ids, rule_probs)]
            for rule_ids, rule_probs in get_sorted_ids_and_probs(
                self.model.forward(mol_outputs), k=top_k
            )
        ]

    def _get_reactions(
        self, inputs: list[Molecule], num_results: int
    ) -> list[Sequence[SingleProductReaction]]:
        return [
            prediction_list[:num_results]
            for prediction_list in self.predict(inputs, top_k=num_results)
        ]



"""Script for augmenting reaction data with the approach used by the R-SMILES model.

This script is a modified version of `preprocessing/generate_PtoR_data.py` from
https://github.com/otori-bird/retrosynthesis/. Some of the utilities are imported from the original
script, but the main logic is cleaned up and simplified by removing features we do not need.
"""

import json
import os
import random
import re
from dataclasses import dataclass, field
from typing import Optional

from omegaconf import MISSING
from rdkit import Chem, RDLogger
from syntheseus.reaction_prediction.data.dataset import DataFold
from syntheseus.reaction_prediction.utils.config import get_config as cli_get_config
from tqdm import tqdm

from retrochimera.utils.logging import get_logger
from retrochimera.utils.root_aligned import get_product_roots

RDLogger.DisableLog("rdApp.*")

logger = get_logger(__name__)


@dataclass
class AugmentRSMILESConfig:
    data_dir: str = MISSING  # Directory containing the data
    augmented_data_dir: str = MISSING  # Directory for saving the augmented data
    folds: list[DataFold] = field(
        default_factory=lambda: [DataFold.TRAIN, DataFold.VALIDATION]
    )  # Folds to augment
    augmentation: int = 1  # Number of augmentations per reaction
    seed: int = 33  # Random seed for reproducibility
    use_random_smiles: bool = False  # If set, use random SMILES instead of root-aligned


def multi_process(data: dict):
    """Multi-process function to process the data."""
    from root_aligned.preprocessing.generate_PtoR_data import (
        clear_map_canonical_smiles,
        get_cano_map_number,
        get_root_id,
    )

    reaction = data["reaction"]
    augmentation = data["augmentation"]

    reactant, _, product = reaction["mapped_reaction_smiles"].split(">")
    pro_mol = Chem.MolFromSmiles(product)
    rea_mol = Chem.MolFromSmiles(reactant)

    # Check data quality.
    return_status = {
        "status": 0,
        "src_smi": [],
        "tgt_smi": [],
        "reaction": reaction,
    }

    if "" == product:
        return_status["status"] = "empty_p"
    if "" == reactant:
        return_status["status"] = "empty_r"
    if rea_mol is None:
        return_status["status"] = "invalid_r"
    if len(rea_mol.GetAtoms()) < 5:
        return_status["status"] = "small_r"
    if pro_mol is None:
        return_status["status"] = "invalid_p"
    if len(pro_mol.GetAtoms()) == 1:
        return_status["status"] = "small_p"
    if not all([a.HasProp("molAtomMapNumber") for a in pro_mol.GetAtoms()]):
        return_status["status"] = "error_mapping_p"

    if return_status["status"] == 0:
        pro_atom_map_numbers = list(map(int, re.findall(r"(?<=:)\d+", product)))
        reactant = reactant.split(".")
        if not data["use_random_smiles"]:
            product_roots = get_product_roots(
                product_atom_ids=pro_atom_map_numbers, num_augmentations=augmentation
            )

            for pro_root_atom_map in product_roots:
                pro_root = get_root_id(pro_mol, root_map_number=pro_root_atom_map)
                cano_atom_map = get_cano_map_number(product, root=pro_root)
                if cano_atom_map is None:
                    return_status["status"] = "error_mapping"
                    return return_status
                pro_smi = clear_map_canonical_smiles(product, canonical=True, root=pro_root)
                aligned_reactants = []
                aligned_reactants_order = []
                rea_atom_map_numbers = [
                    list(map(int, re.findall(r"(?<=:)\d+", rea))) for rea in reactant
                ]
                used_indices = []

                for i, rea_map_number in enumerate(rea_atom_map_numbers):
                    for j, map_number in enumerate(cano_atom_map):
                        if map_number in rea_map_number:
                            rea_root = get_root_id(
                                Chem.MolFromSmiles(reactant[i]), root_map_number=map_number
                            )
                            rea_smi = clear_map_canonical_smiles(
                                reactant[i], canonical=True, root=rea_root
                            )
                            aligned_reactants.append(rea_smi)
                            aligned_reactants_order.append(j)
                            used_indices.append(i)
                            break
                sorted_reactants = sorted(
                    list(zip(aligned_reactants, aligned_reactants_order)), key=lambda x: x[1]
                )
                aligned_reactants = [item[0] for item in sorted_reactants]
                reactant_smi = ".".join(aligned_reactants)

                # Save the resulted smiles.
                return_status["src_smi"].append(pro_smi)
                return_status["tgt_smi"].append(reactant_smi)

            assert len(return_status["src_smi"]) == data["augmentation"]
        else:
            cano_product = clear_map_canonical_smiles(product)
            cano_reactanct = ".".join(
                [
                    clear_map_canonical_smiles(rea)
                    for rea in reactant
                    if set(map(int, re.findall(r"(?<=:)\d+", rea))) & set(pro_atom_map_numbers)
                ]
            )
            return_status["src_smi"].append(cano_product)
            return_status["tgt_smi"].append(cano_reactanct)

            pro_mol = Chem.MolFromSmiles(cano_product)
            rea_mols = [Chem.MolFromSmiles(rea) for rea in cano_reactanct.split(".")]

            for i in range(int(augmentation - 1)):
                pro_smi = Chem.MolToSmiles(pro_mol, doRandom=True)
                rea_smi = [Chem.MolToSmiles(rea_mol, doRandom=True) for rea_mol in rea_mols]
                rea_smi = ".".join(rea_smi)
                return_status["src_smi"].append(pro_smi)
                return_status["tgt_smi"].append(rea_smi)

    return return_status


def preprocess(
    augmented_data_dir: str,
    reaction_list: list,
    fold: DataFold,
    augmentation: int = 1,
    use_random_smiles: bool = False,
) -> None:
    """Preprocess reaction data to extract graph adjacency matrix and features."""

    if not os.path.exists(augmented_data_dir):
        os.makedirs(augmented_data_dir)

    data = [
        {
            "reaction": reaction_list[i],
            "augmentation": augmentation,
            "use_random_smiles": use_random_smiles,
        }
        for i in range(len(reaction_list))
    ]

    skip_dict = {
        "invalid_p": 0,
        "invalid_r": 0,
        "small_p": 0,
        "small_r": 0,
        "error_mapping": 0,
        "error_mapping_p": 0,
        "empty_p": 0,
        "empty_r": 0,
    }

    results = [multi_process(sample) for sample in tqdm(data)]

    with open(os.path.join(augmented_data_dir, f"{fold.value}.jsonl"), "w") as f:
        for result in tqdm(results):
            if result["status"] != 0:
                skip_dict[result["status"]] += 1
                continue

            for src, tgt in zip(result["src_smi"], result["tgt_smi"]):
                reactants_smiles = []
                for reactant in tgt.split("."):
                    reactants_smiles.append({"smiles": reactant})
                assert "." not in src
                products_smiles = [{"smiles": src}]

                updated_reaction = result["reaction"]
                updated_reaction["reactants"] = reactants_smiles
                updated_reaction["products"] = products_smiles
                f.write("{}\n".format(json.dumps(updated_reaction)))

    for key, value in skip_dict.items():
        logger.info(f"{value} samples marked as {key} ({100. * value / len(reaction_list)}%)")


def run_from_config(config: AugmentRSMILESConfig) -> None:
    logger.info(f"Preprocessing dataset from {config.data_dir}")

    random.seed(config.seed)

    data_dir = config.data_dir
    augmented_data_dir = config.augmented_data_dir

    if not os.path.exists(augmented_data_dir):
        os.makedirs(augmented_data_dir)

    for fold in config.folds:
        logger.info(f"Processing fold {fold}")

        with open(os.path.join(data_dir, f"{fold.value}.jsonl"), "r") as f:
            reaction_list = [json.loads(line) for line in f.readlines()]

        logger.info(f"Fold size: {len(reaction_list)}")

        product_smarts_list = list(
            map(lambda x: x["mapped_reaction_smiles"].split(">")[2], reaction_list)
        )

        augmented_data_fold_dir = os.path.join(augmented_data_dir, fold.value)

        # Prevent reactions with multiple products.
        multiple_product_indices = [
            i for i in range(len(product_smarts_list)) if "." in product_smarts_list[i]
        ]
        assert len(multiple_product_indices) == 0, "Multiple products are not supported."

        preprocess(
            augmented_data_fold_dir,
            reaction_list,
            fold,
            config.augmentation,
            use_random_smiles=config.use_random_smiles,
        )


def main(argv: Optional[list[str]]) -> None:
    config: AugmentRSMILESConfig = cli_get_config(argv=argv, config_cls=AugmentRSMILESConfig)
    run_from_config(config)


if __name__ == "__main__":
    main(argv=None)



"""
MIT License

Copyright (c) 2017-Present OpenNMT

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Original Source: https://github.com/OpenNMT/OpenNMT-py/blob/v3.5.1/onmt/modules/position_ffn.py
Modifications:
1. Removed the `self.parallel_gpu` attribute and corresponding logic from the `PositionwiseFeedForward` class, including calls to `torch.distributed.all_reduce` and `torch.nn.utils.skip_init`.
"""
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.checkpoint import checkpoint

from retrochimera.opennmt.modules.rmsnorm import RMSNorm


class ActivationFunction(object):
    relu = "relu"
    gelu = "gelu"
    silu = "silu"
    gated_gelu = "gated-gelu"


# for silu, see: https://arxiv.org/pdf/2002.05202.pdf
ACTIVATION_FUNCTIONS = {
    ActivationFunction.relu: F.relu,
    ActivationFunction.gelu: F.gelu,
    ActivationFunction.silu: F.silu,
    ActivationFunction.gated_gelu: F.gelu,
}


class PositionwiseFeedForward(nn.Module):
    """A two-layer Feed-Forward-Network with residual layer norm.

    Args:
        d_model (int): the size of input for the first-layer of the FFN.
        d_ff (int): the hidden layer size of the second-layer
            of the FNN.
        dropout (float): dropout probability in :math:`[0, 1)`.
        activation_fn (ActivationFunction): activation function used.
        layer_norm (string): 'standard' or 'rms'
    """

    def __init__(
        self,
        d_model,
        d_ff,
        dropout=0.1,
        activation_fn=ActivationFunction.relu,
        add_ffnbias=True,
        parallel_residual=False,
        layer_norm="standard",
        norm_eps=1e-6,
        use_ckpting=[],
    ):
        super(PositionwiseFeedForward, self).__init__()
        self.w_1 = nn.Linear(d_model, d_ff, bias=add_ffnbias)
        self.w_2 = nn.Linear(d_ff, d_model, bias=add_ffnbias)
        if layer_norm == "standard" and not parallel_residual:
            self.layer_norm = nn.LayerNorm(d_model, eps=norm_eps)
        elif layer_norm == "rms" and not parallel_residual:
            self.layer_norm = RMSNorm(d_model, eps=norm_eps)
        elif not parallel_residual:
            raise ValueError(f"{layer_norm} layer norm type is not supported")
        self.parallel_residual = parallel_residual
        self.dropout_p = dropout
        self.dropout_1 = nn.Dropout(dropout)
        self.dropout_2 = nn.Dropout(dropout)
        self.activation = ACTIVATION_FUNCTIONS[activation_fn]
        if activation_fn == "silu" or activation_fn == "gated-gelu":
            self.w_3 = nn.Linear(d_model, d_ff, bias=add_ffnbias)
        else:
            self.w_3 = None
        self.maybe_ckpt = checkpoint if "ffn" in use_ckpting else lambda f, x: f(x)

    def forward(self, x):
        """Layer definition.

        Args:
            x: ``(batch_size, input_len, model_dim)``

        Returns:
            (FloatTensor): Output ``(batch_size, input_len, model_dim)``.
        """
        if not self.parallel_residual:
            norm_x = self.layer_norm(x)
        else:
            norm_x = x.clone()
        inter = self.maybe_ckpt(self.w_1, norm_x)
        inter = self.activation(inter)
        if self.w_3 is not None:
            inter.mul_(self.maybe_ckpt(self.w_3, norm_x))
        if self.dropout_p > 0:
            inter = self.dropout_1(inter)
        inter = self.maybe_ckpt(self.w_2, inter)
        if self.dropout_p > 0:
            inter = self.dropout_2(inter)

        return inter + x

    def update_dropout(self, dropout):
        self.dropout_1.p = dropout
        self.dropout_2.p = dropout



from typing import Optional

import pytest
from syntheseus.reaction_prediction.data.reaction_sample import ReactionSample

from retrochimera.chem.rules import RuleBase
from retrochimera.data.dataset import DataFold, DiskReactionDataset, ReactionDataset
from retrochimera.data.template_reaction_sample import (
    TemplateApplicationResult,
    TemplateReactionSample,
)


@pytest.mark.parametrize("mapped", [False, True])
@pytest.mark.parametrize("include_templates", [False, True])
def test_save_and_load(tmp_path: str, mapped: bool, include_templates: bool) -> None:
    sample_cls: type[ReactionSample]
    if include_templates:
        sample_cls = TemplateReactionSample
        sample_kwargs = {"template_application_results": [TemplateApplicationResult(template_id=0)]}
    else:
        sample_cls = ReactionSample
        sample_kwargs = {}

    samples = [
        sample_cls.from_reaction_smiles_strict(reaction_smiles, mapped=mapped, **sample_kwargs)
        for reaction_smiles in [
            "O[c:1]1[cH:2][c:3](=[O:4])[nH:5][cH:6][cH:7]1>>[cH:1]1[cH:2][c:3](=[O:4])[nH:5][cH:6][cH:7]1",
            "CC(C)(C)OC(=O)[N:1]1[CH2:2][CH2:3][C@H:4]([F:5])[CH2:6]1>>[NH:1]1[CH2:2][CH2:3][C@H:4]([F:5])[CH2:6]1",
        ]
    ]

    for fold in DataFold:
        DiskReactionDataset.save_samples_to_file(data_dir=tmp_path, fold=fold, samples=samples)  # type: ignore

    if include_templates:
        rulebase = RuleBase()
        rulebase.add_rule("[BrH;D0;+0:1]>>C-C(=O)-[Br;H0;D1;+0:1]")
        rulebase.add_rule("[C:2]-[CH3;D1;+0:1]>>O-[CH2;D2;+0:1]-[C:2]")
        rulebase.save_to_file(dir=tmp_path)

    # Now try to load the data we just saved.
    dataset = DiskReactionDataset(tmp_path, sample_cls=sample_cls)

    for fold in DataFold:
        assert list(dataset[fold]) == samples

    if include_templates:
        assert dataset.rulebase.rules == rulebase.rules
    else:
        with pytest.raises(FileNotFoundError):
            dataset.rulebase


@pytest.mark.parametrize("min_rule_support", [None, 3, 5])
@pytest.mark.parametrize("max_num_rules", [None, 2, 5])
def test_limit_rulebase(
    reaction_dataset_uneven: ReactionDataset,
    tmp_path: str,
    min_rule_support: Optional[int],
    max_num_rules: Optional[int],
) -> None:
    DiskReactionDataset.save_samples_to_file(
        data_dir=tmp_path, fold=DataFold.TRAIN, samples=reaction_dataset_uneven[DataFold.TRAIN]
    )
    reaction_dataset_uneven.rulebase.save_to_file(dir=tmp_path)

    reaction_dataset_limited = DiskReactionDataset(
        data_dir=tmp_path,
        sample_cls=TemplateReactionSample,
        rulebase_min_rule_support=min_rule_support,
        rulebase_max_num_rules=max_num_rules,
    )

    expected_num_rules = 4
    if min_rule_support is not None and min_rule_support > 4:
        expected_num_rules = min(expected_num_rules, 8 - min_rule_support)
    if max_num_rules is not None:
        expected_num_rules = min(expected_num_rules, max_num_rules)

    assert len(reaction_dataset_limited.rulebase) == expected_num_rules

    samples_limited = list(reaction_dataset_limited[DataFold.TRAIN])
    assert len(samples_limited) == reaction_dataset_uneven.get_num_samples(DataFold.TRAIN)
